from __future__ import annotations

import json
from datetime import datetime, timedelta, timezone
from typing import Any

from azure.core.exceptions import ResourceExistsError, ResourceNotFoundError
from azure.servicebus import (
    ServiceBusMessage,
    ServiceBusMessageBatch,
    ServiceBusReceivedMessage,
)
from azure.servicebus.management import (
    QueueProperties,
    QueueRuntimeProperties,
    SubscriptionProperties,
    SubscriptionRuntimeProperties,
    TopicProperties,
    TopicRuntimeProperties,
)
from x8._common.azure_provider import AzureProvider
from x8.core import Context, DataModel, NCall, Operation, Response
from x8.core.exceptions import BadRequestError, ConflictError, NotFoundError

from ._models import (
    DEFAULT_SUBSCRIPTION_NAME,
    MessageBatch,
    MessageItem,
    MessageKey,
    MessageProperties,
    MessagePullConfig,
    MessagePutConfig,
    MessageValueType,
    MessagingMode,
    MessagingOperation,
    QueueConfig,
    QueueInfo,
    SubscriptionConfig,
    SubscriptionInfo,
    TopicConfig,
    TopicInfo,
)
from ._operation_parser import MessagingOperationParser


class AzureServiceBusBase(AzureProvider):
    mode: MessagingMode

    queue: str | None
    topic: str | None
    subscription: str | None
    fully_qualified_namespace: str | None
    nparams: dict[str, Any]

    _client: Any
    _aclient: Any
    _topic_cache: dict[str, AzureServiceBusTopic]
    _atopic_cache: dict[str, AzureServiceBusTopic]
    _mgmt_client: Any
    _amgmt_client: Any
    _mgmt_client_helper: Any
    _amgmt_client_helper: Any
    _op_converter: OperationConverter
    _result_converter: ResultConverter

    def __init__(
        self,
        mode: MessagingMode = MessagingMode.QUEUE,
        queue: str | None = None,
        topic: str | None = None,
        subscription: str | None = None,
        fully_qualified_namespace: str | None = None,
        connection_string: str | None = None,
        credential_type: str | None = "default",
        tenant_id: str | None = None,
        client_id: str | None = None,
        client_secret: str | None = None,
        certificate_path: str | None = None,
        nparams: dict[str, Any] = dict(),
        **kwargs: Any,
    ):
        """Initialize.

        Args:
            mode:
                Service Bus mode ("queue" or "pubsub").
            queue:
                Service Bus queue name.
            topic:
                Service Bus topic name.
            subscription:
                Service Bus subscription name.
            fully_qualified_namespace:
                The fully qualified host name for the Service Bus namespace.
                The namespace format is:
                    `<yournamespace>.servicebus.windows.net`.
            connection_string:
                The connection string of a Service Bus.
            credential_type:
                Azure credential type.
            tenant_id:
                Azure tenant id for client_secret credential type.
            client_id:
                Azure client id for client_secret credential type.
            client_secret:
                Azure client secret for client_secret credential type.
            certificate_path:
                Certificate path for certificate credential type.
            nparams:
                Native parameters to Service Bus client.
        """
        self.mode = mode
        self.queue = queue
        self.topic = topic
        self.subscription = subscription
        self.fully_qualified_namespace = fully_qualified_namespace
        self.connection_string = connection_string
        self.nparams = nparams

        self._client = None
        self._aclient = None
        self._mgmt_client = None
        self._amgmt_client = None
        self._topic_cache = dict()
        self._atopic_cache = dict()
        self._op_converter = OperationConverter()
        self._result_converter = ResultConverter()

        if self.mode == MessagingMode.QUEUE:
            self.topic = self.queue
            self.subscription = DEFAULT_SUBSCRIPTION_NAME
        AzureProvider.__init__(
            self,
            credential_type=credential_type,
            tenant_id=tenant_id,
            client_id=client_id,
            client_secret=client_secret,
            certificate_path=certificate_path,
            **kwargs,
        )

    def __setup__(self, context: Context | None = None) -> None:
        if self._client is not None:
            return

        from azure.servicebus import ServiceBusClient

        (self._credential, self._client) = self._init_credential_client(
            ServiceBusClient,
            self._get_credential,
        )

    async def __asetup__(self, context: Context | None = None) -> None:
        if self._client is not None:
            return

        from azure.servicebus.aio import ServiceBusClient

        (
            self._acredential,
            self._aclient,
        ) = self._init_credential_client(
            ServiceBusClient, self._aget_credential
        )

    def _init_mgmt_client(self) -> None:
        if self._mgmt_client is not None:
            return

        from azure.servicebus.management import ServiceBusAdministrationClient

        self._credential, self._mgmt_client = self._init_credential_client(
            ServiceBusAdministrationClient,
            self._get_credential,
        )

    async def _ainit_mgmt_client(self) -> None:
        if self._amgmt_client is not None:
            return

        from azure.servicebus.aio.management import (
            ServiceBusAdministrationClient,
        )

        (
            self._acredential,
            self._amgmt_client,
        ) = self._init_credential_client(
            ServiceBusAdministrationClient,
            self._aget_credential,
        )

    def _init_credential_client(
        self,
        client_type,
        get_credential,
    ):
        credential = None
        if self.connection_string is not None:
            client = client_type.from_connection_string(
                conn_str=self.connection_string,
                **self.nparams,
            )
        else:
            credential = get_credential()
            client = client_type(
                fully_qualified_namespace=self.fully_qualified_namespace,
                credential=credential,
                **self.nparams,
            )
        return credential, client

    def _get_topic_name(
        self, op_parser: MessagingOperationParser
    ) -> str | None:
        if self.mode == MessagingMode.PUBSUB:
            topic_name = op_parser.get_topic()
            topic_name = topic_name or self.topic or self.__component__.topic
        else:
            topic_name = op_parser.get_queue()
            topic_name = topic_name or self.queue or self.__component__.queue
        return topic_name

    def _get_subscription_name(
        self, op_parser: MessagingOperationParser
    ) -> str | None:
        subscription_name = op_parser.get_subscription()
        subscription_name = (
            subscription_name
            or self.subscription
            or self.__component__.subscription
        )
        return subscription_name

    def _init_sender(self, topic: AzureServiceBusTopic, topic_name: str):
        if topic.sender is None:
            if self.mode == MessagingMode.QUEUE:
                topic.sender = self._client.get_queue_sender(
                    queue_name=topic_name
                )
            elif self.mode == MessagingMode.PUBSUB:
                topic.sender = self._client.get_topic_sender(
                    topic_name=topic_name
                )

    def _ainit_sender(self, topic: AzureServiceBusTopic, topic_name: str):
        if topic.sender is None:
            if self.mode == MessagingMode.QUEUE:
                topic.sender = self._aclient.get_queue_sender(
                    queue_name=topic_name
                )
            elif self.mode == MessagingMode.PUBSUB:
                topic.sender = self._aclient.get_topic_sender(
                    topic_name=topic_name
                )

    def _init_receiver(
        self,
        topic: AzureServiceBusTopic,
        topic_name: str,
        subscription_name: str,
    ):
        if subscription_name not in topic.receivers:
            if self.mode == MessagingMode.QUEUE:
                topic.receivers[subscription_name] = (
                    self._client.get_queue_receiver(queue_name=topic_name)
                )
            elif self.mode == MessagingMode.PUBSUB:
                topic.receivers[subscription_name] = (
                    self._client.get_subscription_receiver(
                        topic_name=topic_name,
                        subscription_name=subscription_name,
                    )
                )
            topic.receiver_helpers[subscription_name] = ReceiverHelper(
                topic.receivers[subscription_name]
            )

    def _ainit_receiver(
        self,
        topic: AzureServiceBusTopic,
        topic_name: str,
        subscription_name: str,
    ):
        if subscription_name not in topic.receivers:
            if self.mode == MessagingMode.QUEUE:
                topic.receivers[subscription_name] = (
                    self._aclient.get_queue_receiver(queue_name=topic_name)
                )
            elif self.mode == MessagingMode.PUBSUB:
                topic.receivers[subscription_name] = (
                    self._aclient.get_subscription_receiver(
                        topic_name=topic_name,
                        subscription_name=subscription_name,
                    )
                )
            topic.receiver_helpers[subscription_name] = AsyncReceiverHelper(
                topic.receivers[subscription_name]
            )

    def _get_topic(
        self, op_parser: MessagingOperationParser
    ) -> AzureServiceBusTopic | None:
        if op_parser.is_resource_op():
            return None
        topic_name = self._get_topic_name(op_parser)
        subscription_name = self._get_subscription_name(op_parser)
        if topic_name is None:
            raise BadRequestError("Topic name must be specified")
        if topic_name in self._topic_cache:
            topic = self._topic_cache[topic_name]
        else:
            topic = AzureServiceBusTopic()
            self._topic_cache[topic_name] = topic
        if op_parser.is_sender_op():
            self._init_sender(topic, topic_name)
        elif op_parser.is_receiver_op():
            if subscription_name is None:
                raise BadRequestError("Subscription name must be specified")
            self._init_receiver(topic, topic_name, subscription_name)
        return topic

    def _aget_topic(
        self, op_parser: MessagingOperationParser
    ) -> AzureServiceBusTopic | None:
        if op_parser.is_resource_op():
            return None
        topic_name = self._get_topic_name(op_parser)
        subscription_name = self._get_subscription_name(op_parser)
        if topic_name is None:
            raise BadRequestError("Topic name must be specified")
        if topic_name in self._atopic_cache:
            topic = self._atopic_cache[topic_name]
        else:
            topic = AzureServiceBusTopic()
            self._atopic_cache[topic_name] = topic
        if op_parser.is_sender_op():
            self._ainit_sender(topic, topic_name)
        elif op_parser.is_receiver_op():
            if subscription_name is None:
                raise BadRequestError("Subscription name must be specified")
            self._ainit_receiver(topic, topic_name, subscription_name)
        return topic

    def __run__(
        self,
        operation: Operation | None = None,
        context: Context | None = None,
        **kwargs,
    ) -> Any:
        self.__setup__(context=context)
        op_parser = MessagingOperationParser(operation)
        topic = None
        mgmt_helper = None
        resource_helper = None
        if op_parser.is_resource_op():
            self._init_mgmt_client()
            mgmt_helper = ManagementHelper(
                self._mgmt_client,
                self._op_converter,
                self._result_converter,
            )
        else:
            topic = self._get_topic(op_parser)
        resource_helper = ResourceHelper(
            self._client,
            self._topic_cache,
            self._mgmt_client,
        )
        ncall = self._get_ncall(
            op_parser,
            topic,
            resource_helper,
            mgmt_helper,
        )
        if ncall is None:
            return super().__run__(
                operation=operation,
                context=context,
                **kwargs,
            )
        nresult = ncall.invoke()
        result = self._convert_nresult(nresult, op_parser)
        return Response(result=result, native=dict(result=nresult, call=ncall))

    async def __arun__(
        self,
        operation: Operation | None = None,
        context: Context | None = None,
        **kwargs,
    ) -> Any:
        await self.__asetup__(context=context)
        op_parser = MessagingOperationParser(operation)
        topic = None
        mgmt_helper = None
        resource_helper = None
        if op_parser.is_resource_op():
            await self._ainit_mgmt_client()
            mgmt_helper = AsyncManagementHelper(
                self._amgmt_client,
                self._op_converter,
                self._result_converter,
            )
        else:
            topic = self._aget_topic(op_parser)
        resource_helper = AsyncResourceHelper(
            self._aclient,
            self._atopic_cache,
            self._amgmt_client,
        )
        ncall = self._get_ncall(
            op_parser,
            topic,
            resource_helper,
            mgmt_helper,
        )
        if ncall is None:
            return await super().__arun__(
                operation=operation,
                context=context,
                **kwargs,
            )
        nresult = await ncall.ainvoke()
        result = self._convert_nresult(nresult, op_parser)
        return Response(result=result, native=dict(result=nresult, call=ncall))

    def _get_ncall(
        self,
        op_parser: MessagingOperationParser,
        topic: AzureServiceBusTopic | None,
        resource_helper: Any,
        mgmt_helper: Any,
    ) -> NCall | None:
        if topic is not None:
            sender = topic.sender
            subscription_name = self._get_subscription_name(op_parser)
            if subscription_name in topic.receivers:
                receiver = topic.receivers[subscription_name]
                receiver_helper = topic.receiver_helpers[subscription_name]
        call = None
        nargs = op_parser.get_nargs()
        op_converter = self._op_converter
        # CREATE QUEUE
        if op_parser.op_equals(MessagingOperation.CREATE_QUEUE):
            args = op_converter.convert_create_queue(
                self._get_topic_name(op_parser),
                op_parser.get_queue_config(),
            )
            where_exists = op_parser.get_where_exists()
            call = NCall(
                mgmt_helper.mgmt_client.create_queue,
                args,
                nargs,
                {
                    ResourceExistsError: (
                        ConflictError if where_exists is False else None
                    )
                },
            )
        # UPDATE QUEUE
        elif op_parser.op_equals(MessagingOperation.UPDATE_QUEUE):
            args = {
                "queue_name": self._get_topic_name(op_parser),
                "config": op_parser.get_queue_config(),
                "nargs": nargs,
            }
            call = NCall(
                mgmt_helper.update_queue,
                args,
                None,
                {ResourceNotFoundError: NotFoundError},
            )
        # DROP QUEUE
        elif op_parser.op_equals(MessagingOperation.DROP_QUEUE):
            args = {"queue_name": self._get_topic_name(op_parser)}
            where_exists = op_parser.get_where_exists()
            call = NCall(
                mgmt_helper.mgmt_client.delete_queue,
                args,
                None,
                {
                    ResourceNotFoundError: (
                        NotFoundError if where_exists is True else None
                    )
                },
            )
        # HAS QUEUE
        elif op_parser.op_equals(MessagingOperation.HAS_QUEUE):
            args = {"queue_name": self._get_topic_name(op_parser)}
            call = NCall(
                mgmt_helper.mgmt_client.get_queue,
                args,
                None,
                {
                    ResourceNotFoundError: None,
                },
            )
        # GET QUEUE
        elif op_parser.op_equals(MessagingOperation.GET_QUEUE):
            args = {
                "queue_name": self._get_topic_name(op_parser),
                "nargs": nargs,
            }
            call = NCall(
                mgmt_helper.get_queue,
                args,
                None,
                {
                    ResourceNotFoundError: NotFoundError,
                },
            )
        # LIST QUEUES
        elif op_parser.op_equals(MessagingOperation.LIST_QUEUES):
            args = {"nargs": nargs}
            call = NCall(mgmt_helper.list_queues, args)
        # CREATE TOPIC
        elif op_parser.op_equals(MessagingOperation.CREATE_TOPIC):
            args = op_converter.convert_create_topic(
                self._get_topic_name(op_parser),
                op_parser.get_topic_config(),
            )
            where_exists = op_parser.get_where_exists()
            call = NCall(
                mgmt_helper.mgmt_client.create_topic,
                args,
                nargs,
                {
                    ResourceExistsError: (
                        ConflictError if where_exists is False else None
                    )
                },
            )
        # UPDATE TOPIC
        elif op_parser.op_equals(MessagingOperation.UPDATE_TOPIC):
            args = {
                "topic_name": self._get_topic_name(op_parser),
                "config": op_parser.get_topic_config(),
                "nargs": nargs,
            }
            call = NCall(
                mgmt_helper.update_topic,
                args,
                None,
                {ResourceNotFoundError: NotFoundError},
            )
        # DROP TOPIC
        elif op_parser.op_equals(MessagingOperation.DROP_TOPIC):
            args = {"topic_name": self._get_topic_name(op_parser)}
            where_exists = op_parser.get_where_exists()
            call = NCall(
                mgmt_helper.mgmt_client.delete_topic,
                args,
                None,
                {
                    ResourceNotFoundError: (
                        NotFoundError if where_exists is True else None
                    )
                },
            )
        # LIST TOPICS
        elif op_parser.op_equals(MessagingOperation.LIST_TOPICS):
            args = {"nargs": nargs}
            call = NCall(mgmt_helper.list_topics, args)
        # HAS TOPIC
        elif op_parser.op_equals(MessagingOperation.HAS_TOPIC):
            args = {"topic_name": self._get_topic_name(op_parser)}
            call = NCall(
                mgmt_helper.mgmt_client.get_topic,
                args,
                None,
                {
                    ResourceNotFoundError: None,
                },
            )
        # GET TOPIC
        elif op_parser.op_equals(MessagingOperation.GET_TOPIC):
            args = {
                "topic_name": self._get_topic_name(op_parser),
                "nargs": nargs,
            }
            call = NCall(
                mgmt_helper.get_topic,
                args,
                None,
                {
                    ResourceNotFoundError: NotFoundError,
                },
            )
        # CREATE SUBSCRIPTION
        elif op_parser.op_equals(MessagingOperation.CREATE_SUBSCRIPTION):
            args = op_converter.convert_create_subscription(
                self._get_topic_name(op_parser),
                self._get_subscription_name(op_parser),
                op_parser.get_subscription_config(),
            )
            where_exists = op_parser.get_where_exists()
            call = NCall(
                mgmt_helper.mgmt_client.create_subscription,
                args,
                nargs,
                {
                    ResourceExistsError: (
                        ConflictError if where_exists is False else None
                    )
                },
            )
        # UPDATE SUBSCRIPTION
        elif op_parser.op_equals(MessagingOperation.UPDATE_SUBSCRIPTION):
            args = {
                "topic_name": self._get_topic_name(op_parser),
                "subscription_name": self._get_subscription_name(op_parser),
                "config": op_parser.get_subscription_config(),
                "nargs": nargs,
            }
            call = NCall(
                mgmt_helper.update_subscription,
                args,
                None,
                {ResourceNotFoundError: NotFoundError},
            )
        # DROP SUBSCRIPTION
        elif op_parser.op_equals(MessagingOperation.DROP_SUBSCRIPTION):
            args = {
                "topic_name": self._get_topic_name(op_parser),
                "subscription_name": self._get_subscription_name(op_parser),
            }
            where_exists = op_parser.get_where_exists()
            call = NCall(
                mgmt_helper.mgmt_client.delete_subscription,
                args,
                None,
                {
                    ResourceNotFoundError: (
                        NotFoundError if where_exists is True else None
                    )
                },
            )
        # LIST SUBSCRIPTIONS
        elif op_parser.op_equals(MessagingOperation.LIST_SUBSCRIPTIONS):
            args = {
                "topic_name": self._get_topic_name(op_parser),
                "nargs": nargs,
            }
            call = NCall(mgmt_helper.list_subscriptions, args)
        # HAS SUBSCRIPTION
        elif op_parser.op_equals(MessagingOperation.HAS_SUBSCRIPTION):
            args = {
                "topic_name": self._get_topic_name(op_parser),
                "subscription_name": self._get_subscription_name(op_parser),
            }
            call = NCall(
                mgmt_helper.mgmt_client.get_subscription,
                args,
                None,
                {
                    ResourceNotFoundError: None,
                },
            )
        # GET SUBSCRIPTION
        elif op_parser.op_equals(MessagingOperation.GET_SUBSCRIPTION):
            args = {
                "topic_name": self._get_topic_name(op_parser),
                "subscription_name": self._get_subscription_name(op_parser),
                "nargs": nargs,
            }
            call = NCall(
                mgmt_helper.get_subscription,
                args,
                None,
                {
                    ResourceNotFoundError: NotFoundError,
                },
            )
        # PUT
        elif op_parser.op_equals(MessagingOperation.PUT):
            args = op_converter.convert_put(
                op_parser.get_value(),
                op_parser.get_metadata(),
                op_parser.get_properties(),
                op_parser.get_put_config(),
            )
            call = NCall(
                sender.send_messages,
                args,
                nargs,
            )
        # BATCH
        elif op_parser.op_equals(MessagingOperation.BATCH):
            args = op_converter.convert_batch(
                op_parser.get_batch(),
            )
            call = NCall(
                sender.send_messages,
                args,
                nargs,
            )
        # PULL
        elif op_parser.op_equals(MessagingOperation.PULL):
            args = op_converter.convert_pull(
                op_parser.get_pull_config(),
            )
            call = NCall(
                receiver.receive_messages,
                args,
                nargs,
            )
        # ACK
        elif op_parser.op_equals(MessagingOperation.ACK):
            args = op_converter.convert_key(op_parser.get_key())
            call = NCall(
                receiver.complete_message,
                args,
                nargs,
            )
        # NACK
        elif op_parser.op_equals(MessagingOperation.NACK):
            args = op_converter.convert_key(op_parser.get_key())
            call = NCall(
                receiver.abandon_message,
                args,
                nargs,
            )
        # EXTEND
        elif op_parser.op_equals(MessagingOperation.EXTEND):
            args = op_converter.convert_key(op_parser.get_key())
            call = NCall(
                receiver.renew_message_lock,
                args,
                nargs,
            )
            pass
        # PURGE
        elif op_parser.op_equals(MessagingOperation.PURGE):
            args = {
                "config": op_parser.get_pull_config(),
                "nargs": nargs,
            }
            call = NCall(
                receiver_helper.purge,
                args,
            )
        # CLOSE
        elif op_parser.op_equals(MessagingOperation.CLOSE):
            args = {
                "topic_name": op_parser.get_topic() or op_parser.get_queue(),
                "nargs": nargs,
            }
            call = NCall(resource_helper.close, args)
        return call

    def _convert_nresult(
        self,
        nresult: Any,
        op_parser: MessagingOperationParser,
    ) -> Any:
        result_converter = self._result_converter
        result: Any = None
        # CREATE QUEUE
        if op_parser.op_equals(MessagingOperation.CREATE_QUEUE):
            result = None
        # UPDATE QUEUE
        elif op_parser.op_equals(MessagingOperation.UPDATE_QUEUE):
            result = None
        # DROP QUEUE
        elif op_parser.op_equals(MessagingOperation.DROP_QUEUE):
            result = None
        # LIST QUEUES
        elif op_parser.op_equals(MessagingOperation.LIST_QUEUES):
            result = nresult
        # HAS QUEUE
        elif op_parser.op_equals(MessagingOperation.HAS_QUEUE):
            result = True if nresult else False
        # GET QUEUE
        elif op_parser.op_equals(MessagingOperation.GET_QUEUE):
            result = nresult
        # CREATE TOPIC
        elif op_parser.op_equals(MessagingOperation.CREATE_TOPIC):
            result = None
        # UPDATE TOPIC
        elif op_parser.op_equals(MessagingOperation.UPDATE_TOPIC):
            result = None
        # DROP TOPIC
        elif op_parser.op_equals(MessagingOperation.DROP_TOPIC):
            result = None
        # LIST TOPICS
        elif op_parser.op_equals(MessagingOperation.LIST_TOPICS):
            result = nresult
        # HAS TOPIC
        elif op_parser.op_equals(MessagingOperation.HAS_TOPIC):
            result = True if nresult else False
        # GET TOPIC
        elif op_parser.op_equals(MessagingOperation.GET_TOPIC):
            result = nresult
        # CREATE SUBSCRIPTION
        elif op_parser.op_equals(MessagingOperation.CREATE_SUBSCRIPTION):
            result = None
        # UPDATE SUBSCRIPTION
        elif op_parser.op_equals(MessagingOperation.UPDATE_SUBSCRIPTION):
            result = None
        # DROP SUBSCRIPTION
        elif op_parser.op_equals(MessagingOperation.DROP_SUBSCRIPTION):
            result = None
        # LIST SUBSCRIPTIONS
        elif op_parser.op_equals(MessagingOperation.LIST_SUBSCRIPTIONS):
            result = nresult
        # HAS SUBSCRIPTION
        elif op_parser.op_equals(MessagingOperation.HAS_SUBSCRIPTION):
            result = True if nresult else False
        # GET SUBSCRIPTION
        elif op_parser.op_equals(MessagingOperation.GET_SUBSCRIPTION):
            result = nresult
        # PUT
        elif op_parser.op_equals(MessagingOperation.PUT):
            result = None
        # BATCH
        elif op_parser.op_equals(MessagingOperation.BATCH):
            result = None
        # PULL
        elif op_parser.op_equals(MessagingOperation.PULL):
            result = result_converter.convert_pull(nresult)
        # ACK
        elif op_parser.op_equals(MessagingOperation.ACK):
            result = None
        # NACK
        elif op_parser.op_equals(MessagingOperation.NACK):
            result = None
        # EXTEND
        elif op_parser.op_equals(MessagingOperation.EXTEND):
            result = MessageItem(key=op_parser.get_key())
        # PURGE
        elif op_parser.op_equals(MessagingOperation.PURGE):
            result = None
        # CLOSE
        elif op_parser.op_equals(MessagingOperation.CLOSE):
            result = None
        return result


class ManagementHelper:
    mgmt_client: Any
    op_converter: OperationConverter
    result_converter: ResultConverter

    def __init__(
        self,
        mgmt_client: Any,
        op_converter: OperationConverter,
        result_converter: ResultConverter,
    ):
        self.mgmt_client = mgmt_client
        self.op_converter = op_converter
        self.result_converter = result_converter

    def update_queue(
        self,
        queue_name: str,
        config: QueueConfig | None,
        nargs: Any,
    ) -> None:
        queue_properties: QueueProperties = NCall(
            self.mgmt_client.get_queue,
            {"queue_name": queue_name},
            nargs,
        ).invoke()
        queue_properties = self.op_converter.convert_update_queue(
            queue_properties, config
        )
        NCall(
            self.mgmt_client.update_queue,
            {"queue": queue_properties},
            nargs,
        ).invoke()

    def get_queue(self, queue_name: str, nargs: Any) -> QueueInfo:
        queue_properties: QueueProperties = NCall(
            self.mgmt_client.get_queue,
            {"queue_name": queue_name},
            nargs,
        ).invoke()
        runtime_queue_properties: QueueRuntimeProperties = NCall(
            self.mgmt_client.get_queue_runtime_properties,
            {"queue_name": queue_name},
            nargs,
        ).invoke()
        return self.result_converter.convert_get_queue(
            queue_properties=queue_properties,
            runtime_queue_properties=runtime_queue_properties,
        )

    def list_queues(self, nargs: Any) -> list[str]:
        queues = NCall(self.mgmt_client.list_queues, None, nargs).invoke()
        return [queue.name for queue in queues]

    def update_topic(
        self,
        topic_name: str,
        config: TopicConfig | None,
        nargs: Any,
    ) -> None:
        topic_properties: TopicProperties = NCall(
            self.mgmt_client.get_topic,
            {"topic_name": topic_name},
            nargs,
        ).invoke()
        topic_properties = self.op_converter.convert_update_topic(
            topic_properties, config
        )
        NCall(
            self.mgmt_client.update_topic,
            {"topic": topic_properties},
            nargs,
        ).invoke()

    def get_topic(self, topic_name: str, nargs: Any) -> TopicInfo:
        topic_properties: TopicProperties = NCall(
            self.mgmt_client.get_topic,
            {"topic_name": topic_name},
            nargs,
        ).invoke()
        runtime_topic_properties: TopicRuntimeProperties = NCall(
            self.mgmt_client.get_topic_runtime_properties,
            {"topic_name": topic_name},
            nargs,
        ).invoke()
        return self.result_converter.convert_get_topic(
            topic_properties=topic_properties,
            runtime_topic_properties=runtime_topic_properties,
        )

    def list_topics(self, nargs: Any) -> list[str]:
        topics = NCall(self.mgmt_client.list_topics, None, nargs).invoke()
        return [topic.name for topic in topics]

    def update_subscription(
        self,
        topic_name: str,
        subscription_name: str,
        config: SubscriptionConfig | None,
        nargs: Any,
    ) -> None:
        subscription_properties: SubscriptionProperties = NCall(
            self.mgmt_client.get_subscription,
            {"topic_name": topic_name, "subscription_name": subscription_name},
            nargs,
        ).invoke()
        subscription_properties = (
            self.op_converter.convert_update_subscription(
                subscription_properties, config
            )
        )
        NCall(
            self.mgmt_client.update_subscription,
            {
                "topic_name": topic_name,
                "subscription": subscription_properties,
            },
            nargs,
        ).invoke()

    def get_subscription(
        self,
        topic_name: str,
        subscription_name: str,
        nargs: Any,
    ) -> SubscriptionInfo:
        subscription_properties: SubscriptionProperties = NCall(
            self.mgmt_client.get_subscription,
            {"topic_name": topic_name, "subscription_name": subscription_name},
            nargs,
        ).invoke()
        runtime_subscription_properties: SubscriptionRuntimeProperties = NCall(
            self.mgmt_client.get_subscription_runtime_properties,
            {"topic_name": topic_name, "subscription_name": subscription_name},
            nargs,
        ).invoke()
        return self.result_converter.convert_get_subscription(
            subscription_properties=subscription_properties,
            runtime_subscription_properties=runtime_subscription_properties,
            topic_name=topic_name,
        )

    def list_subscriptions(self, topic_name: str, nargs: Any) -> list[str]:
        subscriptions = NCall(
            self.mgmt_client.list_subscriptions,
            {"topic_name": topic_name},
            nargs,
        ).invoke()
        return [subscription.name for subscription in subscriptions]


class AsyncManagementHelper:
    mgmt_client: Any
    op_converter: OperationConverter
    result_converter: ResultConverter

    def __init__(
        self,
        mgmt_client: Any,
        op_converter: OperationConverter,
        result_converter: ResultConverter,
    ):
        self.mgmt_client = mgmt_client
        self.op_converter = op_converter
        self.result_converter = result_converter

    async def update_queue(
        self,
        queue_name: str,
        config: QueueConfig | None,
        nargs: Any,
    ) -> None:
        queue_properties: QueueProperties = await NCall(
            self.mgmt_client.get_queue,
            {"queue_name": queue_name},
            nargs,
        ).ainvoke()
        queue_properties = self.op_converter.convert_update_queue(
            queue_properties, config
        )
        await NCall(
            self.mgmt_client.update_queue,
            {"queue": queue_properties},
            nargs,
        ).ainvoke()

    async def get_queue(self, queue_name: str, nargs: Any) -> QueueInfo:
        queue_properties: QueueProperties = await NCall(
            self.mgmt_client.get_queue,
            {"queue_name": queue_name},
            nargs,
        ).ainvoke()
        runtime_queue_properties: QueueRuntimeProperties = await NCall(
            self.mgmt_client.get_queue_runtime_properties,
            {"queue_name": queue_name},
            nargs,
        ).ainvoke()
        return self.result_converter.convert_get_queue(
            queue_properties=queue_properties,
            runtime_queue_properties=runtime_queue_properties,
        )

    async def list_queues(self, nargs: Any) -> list[str]:
        queues = NCall(self.mgmt_client.list_queues, None, nargs).invoke()
        return [queue.name async for queue in queues]

    async def update_topic(
        self,
        topic_name: str,
        config: TopicConfig | None,
        nargs: Any,
    ) -> None:
        topic_properties: TopicProperties = await NCall(
            self.mgmt_client.get_topic,
            {"topic_name": topic_name},
            nargs,
        ).ainvoke()
        topic_properties = self.op_converter.convert_update_topic(
            topic_properties, config
        )
        await NCall(
            self.mgmt_client.update_topic,
            {"topic": topic_properties},
            nargs,
        ).ainvoke()

    async def get_topic(self, topic_name: str, nargs: Any) -> TopicInfo:
        topic_properties: TopicProperties = await NCall(
            self.mgmt_client.get_topic,
            {"topic_name": topic_name},
            nargs,
        ).ainvoke()
        runtime_topic_properties: TopicRuntimeProperties = await NCall(
            self.mgmt_client.get_topic_runtime_properties,
            {"topic_name": topic_name},
            nargs,
        ).ainvoke()
        return self.result_converter.convert_get_topic(
            topic_properties=topic_properties,
            runtime_topic_properties=runtime_topic_properties,
        )

    async def list_topics(self, nargs: Any) -> list[str]:
        topics = NCall(self.mgmt_client.list_topics, None, nargs).invoke()
        return [topic.name async for topic in topics]

    async def update_subscription(
        self,
        topic_name: str,
        subscription_name: str,
        config: SubscriptionConfig | None,
        nargs: Any,
    ) -> None:
        subscription_properties: SubscriptionProperties = await NCall(
            self.mgmt_client.get_subscription,
            {"topic_name": topic_name, "subscription_name": subscription_name},
            nargs,
        ).ainvoke()
        subscription_properties = (
            self.op_converter.convert_update_subscription(
                subscription_properties, config
            )
        )
        await NCall(
            self.mgmt_client.update_subscription,
            {
                "topic_name": topic_name,
                "subscription": subscription_properties,
            },
            nargs,
        ).ainvoke()

    async def get_subscription(
        self,
        topic_name: str,
        subscription_name: str,
        nargs: Any,
    ) -> SubscriptionInfo:
        subscription_properties: SubscriptionProperties = await NCall(
            self.mgmt_client.get_subscription,
            {"topic_name": topic_name, "subscription_name": subscription_name},
            nargs,
        ).ainvoke()
        runtime_subscription_properties: SubscriptionRuntimeProperties = (
            await NCall(
                self.mgmt_client.get_subscription_runtime_properties,
                {
                    "topic_name": topic_name,
                    "subscription_name": subscription_name,
                },
                nargs,
            ).ainvoke()
        )
        return self.result_converter.convert_get_subscription(
            subscription_properties=subscription_properties,
            runtime_subscription_properties=runtime_subscription_properties,
            topic_name=topic_name,
        )

    async def list_subscriptions(
        self, topic_name: str, nargs: Any
    ) -> list[str]:
        subscriptions = NCall(
            self.mgmt_client.list_subscriptions,
            {"topic_name": topic_name},
            nargs,
        ).invoke()
        return [subscription.name async for subscription in subscriptions]


class ResourceHelper:
    client: Any
    topic_cache: dict[str, AzureServiceBusTopic]
    mgmt_client: Any

    def __init__(self, client: Any, topic_cache: dict, mgmt_client: Any):
        self.client = client
        self.topic_cache = topic_cache

    def close(self, topic_name: str | None, nargs: Any) -> Any:
        if topic_name is not None and topic_name in self.topic_cache:
            self.topic_cache.pop(topic_name)
        else:
            self.topic_cache.clear()


class AsyncResourceHelper:
    client: Any
    topic_cache: dict[str, AzureServiceBusTopic]
    mgmt_client: Any

    def __init__(self, client: Any, topic_cache: dict, mgmt_client: Any):
        self.client = client
        self.topic_cache = topic_cache
        self.mgmt_client = mgmt_client

    async def close(self, topic_name: str | None, nargs: Any) -> None:
        if topic_name is not None and topic_name in self.topic_cache:
            topic = self.topic_cache[topic_name]
            if topic.sender is not None:
                await NCall(
                    topic.sender.close,
                    None,
                    nargs,
                ).ainvoke()
            for receiver in topic.receivers.values():
                await NCall(
                    receiver.close,
                    None,
                    nargs,
                ).ainvoke()
            self.topic_cache.pop(topic_name)
        else:
            for key, topic in self.topic_cache.items():
                if topic.sender is not None:
                    await NCall(
                        topic.sender.close,
                        None,
                        nargs,
                    ).ainvoke()
                for receiver in topic.receivers.values():
                    await NCall(
                        receiver.close,
                        None,
                        nargs,
                    ).ainvoke()
            await NCall(
                self.client.close,
                None,
                nargs,
            ).ainvoke()
            self.topic_cache.clear()
        if self.mgmt_client is not None:
            await NCall(
                self.mgmt_client.close,
                None,
                nargs,
            ).ainvoke()


class ReceiverHelper:
    receiver: Any

    def __init__(self, receiver: Any):
        self.receiver = receiver

    def purge(
        self,
        config: MessagePullConfig | None,
        nargs: Any,
    ) -> None:
        while True:
            max_message_count = 10
            max_wait_time = None
            if config is not None and config.max_count is not None:
                max_message_count = config.max_count
            if config is not None and config.max_wait_time is not None:
                max_wait_time = config.max_wait_time
            messages = NCall(
                self.receiver.receive_messages,
                {
                    "max_message_count": max_message_count,
                    "max_wait_time": max_wait_time,
                },
                nargs,
            ).invoke()
            if len(messages) == 0:
                break
            for message in messages:
                NCall(
                    self.receiver.complete_message, {"message": message}
                ).invoke()


class AsyncReceiverHelper:
    receiver: Any

    def __init__(self, receiver: Any):
        self.receiver = receiver

    async def purge(
        self,
        config: MessagePullConfig | None,
        nargs: Any,
    ) -> None:
        while True:
            max_message_count = 10
            max_wait_time = None
            if config is not None and config.max_count is not None:
                max_message_count = config.max_count
            if config is not None and config.max_wait_time is not None:
                max_wait_time = config.max_wait_time
            messages = await NCall(
                self.receiver.receive_messages,
                {
                    "max_message_count": max_message_count,
                    "max_wait_time": max_wait_time,
                },
                nargs,
            ).ainvoke()
            if len(messages) == 0:
                break
            tasks = []
            for message in messages:
                tasks.append(
                    NCall(
                        self.receiver.complete_message, {"message": message}
                    ).invoke(),
                )
            import asyncio

            await asyncio.gather(*tasks)


class OperationConverter:
    def convert_update_queue(
        self,
        queue_properties: QueueProperties,
        config: QueueConfig | None,
    ) -> QueueProperties:
        if config is not None:
            if config.visibility_timeout is not None:
                queue_properties.lock_duration = timedelta(
                    seconds=config.visibility_timeout
                )
            if config.ttl is not None:
                queue_properties.default_message_time_to_live = timedelta(
                    seconds=config.ttl
                )
            if config.max_delivery_count is not None:
                queue_properties.max_delivery_count = config.max_delivery_count

            if config.dlq_nref is not None:
                queue_properties.forward_dead_lettered_messages_to = (
                    config.dlq_nref
                )
            if config.nconfig:
                for key, value in config.nconfig.items():
                    if hasattr(queue_properties, key):
                        setattr(queue_properties, key, value)
        return queue_properties

    def convert_update_topic(
        self,
        topic_properties: TopicProperties,
        config: TopicConfig | None,
    ) -> TopicProperties:
        if config is not None:
            if config.ttl is not None:
                topic_properties.default_message_time_to_live = timedelta(
                    seconds=config.ttl
                )
            if config.nconfig:
                for key, value in config.nconfig.items():
                    if hasattr(topic_properties, key):
                        setattr(topic_properties, key, value)
        return topic_properties

    def convert_update_subscription(
        self,
        subscription_properties: SubscriptionProperties,
        config: SubscriptionConfig | None,
    ) -> SubscriptionProperties:
        if config is not None:
            if config.visibility_timeout is not None:
                subscription_properties.lock_duration = timedelta(
                    seconds=config.visibility_timeout
                )
            if config.ttl is not None:
                subscription_properties.default_message_time_to_live = (
                    timedelta(seconds=config.ttl)
                )
            if config.max_delivery_count is not None:
                subscription_properties.max_delivery_count = (
                    config.max_delivery_count
                )

            if config.dlq_nref is not None:
                subscription_properties.forward_dead_lettered_messages_to = (
                    config.dlq_nref
                )
            if config.nconfig:
                for key, value in config.nconfig.items():
                    if hasattr(subscription_properties, key):
                        setattr(subscription_properties, key, value)
        return subscription_properties

    def _convert_queue_config(
        self,
        config: QueueConfig | None,
    ) -> dict:
        args: dict = {}

        if config:
            if config.visibility_timeout is not None:
                args["lock_duration"] = timedelta(
                    seconds=config.visibility_timeout
                )
            if config.ttl is not None:
                args["default_message_time_to_live"] = timedelta(
                    seconds=config.ttl
                )
            if config.max_delivery_count is not None:
                args["max_delivery_count"] = config.max_delivery_count

            if config.dlq_nref is not None:
                args["forward_dead_lettered_messages_to"] = config.dlq_nref

            if config.nconfig:
                args.update(config.nconfig)

        return args

    def _convert_topic_config(
        self,
        config: TopicConfig | None,
    ) -> dict:
        args: dict = {}
        if config:
            if config.ttl is not None:
                args["default_message_time_to_live"] = timedelta(
                    seconds=config.ttl
                )
            if config.nconfig:
                args.update(config.nconfig)
        return args

    def _convert_subscription_config(
        self,
        config: SubscriptionConfig | None,
    ) -> dict:
        args: dict = {}
        if config:
            if config.visibility_timeout is not None:
                args["lock_duration"] = timedelta(
                    seconds=config.visibility_timeout
                )
            if config.ttl is not None:
                args["default_message_time_to_live"] = timedelta(
                    seconds=config.ttl
                )
            if config.max_delivery_count is not None:
                args["max_delivery_count"] = config.max_delivery_count

            if config.dlq_nref is not None:
                args["forward_dead_lettered_messages_to"] = config.dlq_nref

            if config.nconfig:
                args.update(config.nconfig)
        return args

    def convert_create_queue(
        self,
        queue_name: str | None,
        config: QueueConfig | None,
    ):
        args: dict = {
            "queue_name": queue_name,
        }
        if config:
            args.update(self._convert_queue_config(config))
        return args

    def convert_create_topic(
        self,
        topic_name: str | None,
        config: TopicConfig | None,
    ):
        args: dict = {
            "topic_name": topic_name,
        }
        if config:
            args.update(self._convert_topic_config(config))
        return args

    def convert_create_subscription(
        self,
        topic_name: str | None,
        subscription_name: str | None,
        config: SubscriptionConfig | None,
    ):
        args: dict = {
            "topic_name": topic_name,
            "subscription_name": subscription_name,
        }
        if config:
            args.update(self._convert_subscription_config(config))
        return args

    def convert_put(
        self,
        value: MessageValueType,
        metadata: dict | None,
        properties: MessageProperties | None,
        config: MessagePutConfig | None,
    ) -> dict:
        message_id = None
        content_type = None
        scheduled_time: datetime | None = None
        group_id = None
        body: str | bytes | None = None
        if properties is not None:
            message_id = properties.message_id
            content_type = properties.content_type
            group_id = properties.group_id
        if isinstance(value, str):
            body = value
            content_type = "text/plain"
        elif isinstance(value, bytes):
            body = value
            content_type = "application/octet-stream"
        elif isinstance(value, dict):
            body = json.dumps(value)
            content_type = "application/json"
        elif isinstance(value, DataModel):
            body = value.to_json()
            content_type = "application/json"
        else:
            raise BadRequestError("Message type not supported")
        if config is not None and config.delay is not None:
            scheduled_time = datetime.now(timezone.utc) + timedelta(
                seconds=config.delay
            )
        return {
            "message": ServiceBusMessage(
                body=body,
                message_id=message_id,
                content_type=content_type,
                session_id=group_id,
                scheduled_enqueue_time_utc=scheduled_time,
                application_properties=metadata,
            )
        }

    def convert_batch(self, batch: MessageBatch) -> dict:
        sb_batch = ServiceBusMessageBatch()
        for operation in batch.operations:
            op_parser = MessagingOperationParser(operation)
            sb_batch.add_message(
                message=self.convert_put(
                    op_parser.get_value(),
                    op_parser.get_metadata(),
                    op_parser.get_properties(),
                    op_parser.get_put_config(),
                )["message"]
            )
        return {"message": sb_batch}

    def convert_pull(self, config: MessagePullConfig | None) -> dict:
        max_message_count = 1
        max_wait_time = None
        if config is not None and config.max_count is not None:
            max_message_count = config.max_count
        if config is not None and config.max_wait_time is not None:
            max_wait_time = config.max_wait_time
        return {
            "max_message_count": max_message_count,
            "max_wait_time": max_wait_time,
        }

    def convert_key(self, key: MessageKey) -> dict:
        return {"message": key.nref}


class ResultConverter:
    def _convert_key(self, message: ServiceBusReceivedMessage) -> MessageKey:
        return MessageKey(nref=message)

    def _convert_value(
        self, message: ServiceBusReceivedMessage
    ) -> MessageValueType:
        if message.content_type == "application/octet-stream":
            return b"".join(message.raw_amqp_message.body)
        elif message.content_type == "application/json":
            return json.loads(str(message))
        return str(message)

    def _convert_metadata(
        self, message: ServiceBusReceivedMessage
    ) -> dict | None:
        if message.application_properties is None:
            return None
        metadata = {}
        for key, value in message.application_properties.items():
            if isinstance(key, bytes):
                k = key.decode("utf-8")
            else:
                k = key
            if isinstance(value, bytes):
                v: Any = value.decode("utf-8")
            else:
                v = value
            metadata[k] = v
        return metadata

    def _convert_properties(
        self,
        message: ServiceBusReceivedMessage,
    ) -> MessageProperties:
        return MessageProperties(
            message_id=message.message_id,
            content_type=message.content_type,
            group_id=message.session_id,
            enqueued_time=(
                message.enqueued_time_utc.timestamp()
                if message.enqueued_time_utc
                else None
            ),
            delivery_count=message.delivery_count,
        )

    def _convert_message(
        self,
        message: ServiceBusReceivedMessage,
    ) -> MessageItem:
        return MessageItem(
            key=self._convert_key(message=message),
            value=self._convert_value(message=message),
            metadata=self._convert_metadata(message=message),
            properties=self._convert_properties(message=message),
        )

    def convert_pull(
        self, messages: list[ServiceBusReceivedMessage]
    ) -> list[MessageItem]:
        result = []
        for message in messages:
            result.append(self._convert_message(message))
        return result

    def convert_get_queue(
        self,
        queue_properties: QueueProperties,
        runtime_queue_properties: QueueRuntimeProperties,
    ) -> QueueInfo:
        rqp = runtime_queue_properties
        visibility_timeout = None
        ttl = None
        qp_ld = queue_properties.lock_duration
        if qp_ld is not None:
            if isinstance(qp_ld, timedelta):
                visibility_timeout = qp_ld.total_seconds()
            else:
                visibility_timeout = None
        qp_ttl = queue_properties.default_message_time_to_live
        if qp_ttl is not None:
            if isinstance(qp_ttl, timedelta):
                ttl = qp_ttl.total_seconds()
            else:
                ttl = None
        config = QueueConfig(
            visibility_timeout=visibility_timeout,
            ttl=ttl,
            max_delivery_count=queue_properties.max_delivery_count,
            dlq_nref=queue_properties.forward_dead_lettered_messages_to,
            nconfig=queue_properties.__dict__,
        )
        queue_info = QueueInfo(
            name=queue_properties.name,
            active_message_count=rqp.active_message_count,
            inflight_message_count=(
                (rqp.total_message_count or 0)
                - (rqp.active_message_count or 0)
            ),
            scheduled_message_count=rqp.scheduled_message_count,
            config=config,
            nref=queue_properties.name,
        )
        return queue_info

    def convert_get_topic(
        self,
        topic_properties: TopicProperties,
        runtime_topic_properties: TopicRuntimeProperties,
    ) -> TopicInfo:
        rtp = runtime_topic_properties
        ttl = None
        tp_ttl = topic_properties.default_message_time_to_live
        if tp_ttl is not None:
            if isinstance(tp_ttl, timedelta):
                ttl = tp_ttl.total_seconds()
            else:
                ttl = None
        config = TopicConfig(
            ttl=ttl,
            nconfig=topic_properties.__dict__,
        )
        topic_info = TopicInfo(
            name=topic_properties.name,
            subscription_count=rtp.subscription_count,
            scheduled_message_count=rtp.scheduled_message_count,
            config=config,
            nref=topic_properties.name,
        )
        return topic_info

    def convert_get_subscription(
        self,
        subscription_properties: SubscriptionProperties,
        runtime_subscription_properties: SubscriptionRuntimeProperties,
        topic_name: str,
    ) -> SubscriptionInfo:
        rsp = runtime_subscription_properties
        visibility_timeout = None
        ttl = None
        sp_ld = subscription_properties.lock_duration
        if sp_ld is not None:
            if isinstance(sp_ld, timedelta):
                visibility_timeout = sp_ld.total_seconds()
            else:
                visibility_timeout = None
        sp_ttl = subscription_properties.default_message_time_to_live
        if sp_ttl is not None:
            if isinstance(sp_ttl, timedelta):
                ttl = sp_ttl.total_seconds()
            else:
                ttl = None
        config = SubscriptionConfig(
            visibility_timeout=visibility_timeout,
            ttl=ttl,
            max_delivery_count=subscription_properties.max_delivery_count,
            dlq_nref=subscription_properties.forward_dead_lettered_messages_to,
            nconfig=subscription_properties.__dict__,
        )
        subscription_info = SubscriptionInfo(
            name=subscription_properties.name,
            topic=topic_name,
            active_message_count=rsp.active_message_count,
            inflight_message_count=(
                (rsp.total_message_count or 0)
                - (rsp.active_message_count or 0)
            ),
            config=config,
            nref=subscription_properties.name,
        )
        return subscription_info


class AzureServiceBusTopic:
    sender: Any
    receivers: dict[str, Any]
    receiver_helpers: dict[str, Any]

    def __init__(self):
        self.sender = None
        self.receivers = {}
        self.receiver_helpers = {}


"""
    This is a bug in the client. The last message will not be received.
    servicebus_client = ServiceBusClient.from_connection_string(
        conn_str=NAMESPACE_CONNECTION_STR,
        logging_enable=True)
    sender = servicebus_client.get_queue_sender(queue_name=QUEUE_NAME)
    receiver = servicebus_client.get_queue_receiver(queue_name=QUEUE_NAME)
    await sender.send_messages(ServiceBusMessage("Message 1"))
    while True:
        messages = await receiver.receive_messages(max_wait_time=2, max_message_count=10)
        if len(messages) == 0:
            break
        for message in messages:
            print(f"Completing - {message.message}")
            await receiver.complete_message(message)
    await sender.send_messages(ServiceBusMessage("Message 2"))
    messages = await receiver.receive_messages(max_message_count=1)
    await receiver.complete_message(messages[0])
    print(f"Completing - {messages[0].message}")
    await sender.send_messages(ServiceBusMessage("Message 3"))
    messages = await receiver.receive_messages(max_message_count=1)
    await receiver.complete_message(messages[0])
    print(f"Completing - {messages[0].message}")
    await sender.send_messages(ServiceBusMessage("Message 4"))
    messages = await receiver.receive_messages(max_message_count=1)
    await receiver.complete_message(messages[0])
    print(f"Completing - {messages[0].message}")
    await servicebus_client.close()
"""  # noqa



import hashlib
import re

from ._models import (
    ArrayIndex,
    AscIndex,
    CompositeIndex,
    DescIndex,
    ExcludeIndex,
    FieldIndex,
    GeospatialIndex,
    HashIndex,
    Index,
    IndexStatus,
    RangeIndex,
    SparseVectorIndex,
    TextIndex,
    VectorIndex,
    WildcardIndex,
)


class IndexHelper:
    @staticmethod
    def check_index_status(
        indexes: list, index: Index
    ) -> tuple[IndexStatus, Index | None]:
        matched_index = IndexHelper.match_index(indexes, index, False)
        if matched_index:
            return IndexStatus.EXISTS, matched_index
        matched_index = IndexHelper.match_index(indexes, index, True)
        if matched_index:
            return IndexStatus.COVERED, matched_index
        return IndexStatus.NOT_EXISTS, None

    @staticmethod
    def get_composite_types_from_name(name: str) -> list[str]:
        splits = name.split("_")
        if (
            len(splits) > 3
            and splits[1] == "composite"
            and splits[2].isnumeric()
        ):
            n = int(splits[2])
            start = 3
            end = 3 + n
            return splits[start:end]
        return []

    @staticmethod
    def get_type_from_name(name: str) -> str | None:
        splits = name.split("_")
        if len(splits) > 1:
            return splits[1]
        return None

    @staticmethod
    def hash_fields(fields: list[str]):
        concatenated = ",".join(fields)
        hash_object = hashlib.sha256(concatenated.encode("utf-8"))
        short_hash = hash_object.hexdigest()[:8]
        return short_hash

    @staticmethod
    def is_index(
        index: Index,
        match_index: Index,
        superset_match: bool = False,
    ) -> bool:
        if superset_match:
            if index.type == "asc" and match_index.type == "hash":
                pass
            elif index.type == "desc" and match_index.type == "hash":
                pass
            elif index.type == "range" and match_index.type == "asc":
                pass
            elif index.type == "range" and match_index.type == "desc":
                pass
            elif index.type == "field" and match_index.type == "hash":
                pass
            elif index.type == "field" and match_index.type == "asc":
                pass
            elif index.type == "field" and match_index.type == "desc":
                pass
            elif index.type != match_index.type:
                return False
        else:
            if index.type != match_index.type:
                return False
        if isinstance(index, CompositeIndex) and isinstance(
            match_index, CompositeIndex
        ):
            if len(index.fields) != len(match_index.fields):
                return False
            for a, b in zip(index.fields, match_index.fields):
                if not IndexHelper.is_index(a, b, superset_match):
                    return False
            return True
        else:
            if getattr(index, "field") != getattr(match_index, "field"):
                return False
            if (index, WildcardIndex) and isinstance(
                match_index, WildcardIndex
            ):
                if sorted(getattr(index, "excluded")) != sorted(
                    getattr(match_index, "excluded")
                ):
                    return False
        return True

    @staticmethod
    def match_index(
        indexes: list[Index],
        match_index: Index,
        superset_match: bool = False,
    ) -> Index | None:
        for index in indexes:
            if IndexHelper.is_index(index, match_index, superset_match):
                return index
        return None

    @staticmethod
    def convert_index_name(
        index: Index,
        collection: str | None = None,
        prefix: str | None = "idx",
    ) -> str:
        if index.name:
            return index.name
        if isinstance(
            index,
            (
                RangeIndex,
                HashIndex,
                FieldIndex,
                AscIndex,
                DescIndex,
                GeospatialIndex,
                VectorIndex,
                TextIndex,
                ArrayIndex,
                SparseVectorIndex,
                ExcludeIndex,
            ),
        ):
            field = index.field.replace(".", "_")
            field = re.sub(r"\[\d+\]", "", field)
            name = f"{index.type}_{field}"
        elif isinstance(index, CompositeIndex):
            name = f"{index.type}_{len(index.fields)}"
            for part in index.fields:
                name = f"{name}_{part.type}"
            field_hash = IndexHelper.hash_fields(
                [part.field for part in index.fields]
            )
            name = f"{name}_{field_hash}"
        elif isinstance(index, WildcardIndex):
            field = index.field.replace(".", "_").replace("*", "")
            field = re.sub(r"\[\d+\]", "", field)
            name = f"{index.type}_{field}"
            name = name.rstrip("_")
        else:
            name = "idx"
        if prefix:
            name = f"{prefix}_{name}"
        if collection:
            name = f"{name}_{collection}"
        return name



# type: ignore
import time

import pytest

from x8.core.exceptions import ConflictError, NotFoundError
from x8.storage.vector_store import CollectionStatus, VectorBatch, VectorItem

from ._data import vectors
from ._providers import VectorStoreProvider
from ._sync_and_async_client import VectorStoreSyncAndAsyncClient

queries = [
    {
        "args": {
            "search": "vector_search(vector=@p1)",
            "order_by": "$id",
            "params": {"p1": [0.0] * 4},
        },
        "result_index": [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
        "count": 10,
    },
    {
        "args": {
            "search": "vector_search(vector=@p1)",
            "limit": 5,
            "params": {"p1": [0.1, 0.1, 0.1, 0.1]},
        },
        "result_index": [9, 8, 7, 6, 5],
        "count": 5,
    },
    {
        "args": {
            "search": "vector_search(vector=@p1)",
            "limit": 5,
            "offset": 2,
            "params": {"p1": [0.1, 0.1, 0.1, 0.1]},
        },
        "result_index": [7, 6, 5, 4, 3],
        "count": 5,
    },
    {
        "args": {
            "search": "vector_search(vector=@p1)",
            "order_by": "$score",
            "limit": 5,
            "params": {"p1": [0.1, 0.1, 0.1, 0.1]},
        },
        "result_index": [5, 6, 7, 8, 9],
        "count": 5,
    },
    {
        "args": {
            "search": "vector_search(vector=@p1)",
            "limit": 5,
            "where": "$metadata.int > 80",
            "params": {"p1": [0.1, 0.1, 0.1, 0.1]},
        },
        "result_index": [9, 8, 7],
        "count": 3,
    },
    {
        "args": {
            "search": "vector_search(vector=@p1, sparse_vector=@p2)",
            "limit": 5,
            "params": {"p1": [0.1, 0.1, 0.1, 0.1], "p2": {12: 0.11, 13: 0.22}},
        },
        "result_index": [9, 8, 7, 6, 5],
        "count": 5,
        "except_providers": [
            VectorStoreProvider.MILVUS,
            VectorStoreProvider.CHROMA,
            VectorStoreProvider.WEAVIATE,
        ],
    },
    {
        "args": {
            "where": "$metadata.bool = true",
            "order_by": "$id",
        },
        "result_index": [0, 1, 2, 3, 4],
        "count": 5,
    },
    {
        "args": {
            "select": "*",
            "where": "$metadata.bool = false",
            "order_by": "$id",
        },
        "result_index": [5, 6, 7, 8, 9],
        "count": 5,
    },
    {
        "args": {
            "select": "$metadata",
            "where": "$metadata.bool = true",
            "order_by": "$id",
        },
        "result_index": [0, 1, 2, 3, 4],
        "count": 5,
    },
    {
        "args": {
            "select": "$value, $metadata",
            "where": "$metadata.bool = true",
            "order_by": "$id",
        },
        "result_index": [0, 1, 2, 3, 4],
        "count": 5,
    },
    {
        "args": {
            "where": """$metadata.bool = true
                        and $metadata.int <= 30""",
            "order_by": "$id",
        },
        "result_index": [0, 1, 2],
        "count": 3,
    },
    {
        "args": {
            "where": """$metadata.bool = true
                        and $metadata.int >= 30
                        and $metadata.int <= 50""",
            "order_by": "$id",
        },
        "result_index": [1, 2, 3, 4],
        "count": 4,
    },
    {
        "args": {
            "where": """$metadata.bool = true
                        and $metadata.int between 30 and 50""",
            "order_by": "$id",
        },
        "result_index": [1, 2, 3, 4],
        "count": 4,
    },
    {
        "args": {
            "where": """$metadata.bool = true
                        and $metadata.int < 50
                        and $metadata.str = 'value1'""",
            "order_by": "$id",
        },
        "result_index": [0],
        "count": 1,
    },
    {
        "args": {
            "where": """$metadata.bool = true
                        and $metadata.int < 50
                        and ($metadata.str = 'value1'
                        or $metadata.str = 'value2')""",
            "order_by": "$id",
        },
        "result_index": [0, 1],
        "count": 2,
    },
    {
        "args": {
            "where": """$metadata.bool = true
                        and $metadata.int < 50
                        and $metadata.str in ('value1', 'value2')""",
            "order_by": "$id",
        },
        "result_index": [0, 1],
        "count": 2,
    },
    {
        "args": {
            "where": """$metadata.bool = true
                        and $metadata.int <= 50
                        and $metadata.str not in ('value1', 'value2')""",
            "order_by": "$id",
        },
        "result_index": [2, 3, 4],
        "count": 1,
    },
    {
        "args": {
            "where": """$metadata.bool = true
                        and array_contains($metadata.arr, 'tag1')""",
            "order_by": "$id",
        },
        "result_index": [0, 1],
        "count": 2,
        "except_providers": [VectorStoreProvider.CHROMA],
    },
    {
        "args": {
            "where": """$metadata.bool = true
                        and array_contains_any($metadata.arr,
                        ['tag1', 'tag2'])""",
            "order_by": "$id",
        },
        "result_index": [0, 1, 2, 3],
        "count": 4,
        "except_providers": [VectorStoreProvider.CHROMA],
    },
    {
        "args": {
            "where": """$metadata.int != 30
                        and $metadata.bool = true
                        and is_defined($metadata.str)
                        """,
            "order_by": "$id",
        },
        "result_index": [0, 3, 4],
        "count": 3,
        "except_providers": [
            VectorStoreProvider.CHROMA,
            VectorStoreProvider.WEAVIATE,
        ],
    },
    {
        "args": {
            "where": """$metadata.int != 30
                        and $metadata.bool = true
                        and is_not_defined($metadata.str2)
                        """,
            "order_by": "$id",
        },
        "result_index": [0, 3, 4],
        "count": 3,
        "except_providers": [
            VectorStoreProvider.CHROMA,
            VectorStoreProvider.WEAVIATE,
        ],
    },
]

except_update_providers = [
    VectorStoreProvider.MILVUS,
    VectorStoreProvider.QDRANT,
]
except_sparse_vector_providers = [
    VectorStoreProvider.MILVUS,
    VectorStoreProvider.CHROMA,
    VectorStoreProvider.WEAVIATE,
]
except_delete_filter_providers = [VectorStoreProvider.PINECONE]
except_delete_all_providers = [
    VectorStoreProvider.MILVUS,
    VectorStoreProvider.QDRANT,
    VectorStoreProvider.CHROMA,
    VectorStoreProvider.WEAVIATE,
]
except_count_providers = [VectorStoreProvider.MILVUS]
except_metadata_list_providers = [VectorStoreProvider.CHROMA]


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "provider_type",
    [
        # VectorStoreProvider.PINECONE,
        # VectorStoreProvider.MILVUS,
        # VectorStoreProvider.QDRANT,
        VectorStoreProvider.CHROMA,
        # VectorStoreProvider.WEAVIATE,
    ],
)
@pytest.mark.parametrize(
    "async_call",
    [False, True],
)
async def test_collection(provider_type: str, async_call: bool):
    new_collection = "ntest"
    client = VectorStoreSyncAndAsyncClient(
        provider_type=provider_type, async_call=async_call
    )

    response = await client.list_collections()
    result = response.result
    if new_collection in result:
        await client.drop_collection(new_collection)

    response = await client.list_collections()
    result = response.result
    assert new_collection not in result

    response = await client.has_collection(new_collection)
    result = response.result
    assert result is False

    response = await client.create_collection(collection=new_collection)
    result = response.result
    assert result.status == CollectionStatus.CREATED

    response = await client.create_collection(collection=new_collection)
    result = response.result
    assert result.status == CollectionStatus.EXISTS
    with pytest.raises(ConflictError):
        await client.create_collection(
            collection=new_collection, where="not_exists()"
        )

    item = vectors[0]
    await client.put(
        key=get_key(item),
        value=item["value"],
        metadata=item["metadata"],
        collection=new_collection,
    )
    await wait_for_put(provider_type)
    response = await client.get(get_key(item), collection=new_collection)
    result = response.result
    assert_vector_item(provider_type, result, item)

    response = await client.list_collections()
    result = response.result
    assert new_collection in result

    response = await client.has_collection(new_collection)
    result = response.result
    assert result is True

    response = await client.drop_collection(new_collection)
    result = response.result
    assert result.status == CollectionStatus.DROPPED
    response = await client.list_collections()
    result = response.result
    assert new_collection not in result

    response = await client.drop_collection(
        collection=new_collection,
    )
    result = response.result
    assert result.status == CollectionStatus.NOT_EXISTS
    with pytest.raises(NotFoundError):
        await client.drop_collection(
            collection=new_collection, where="exists()"
        )

    await client.close()


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "provider_type",
    [
        VectorStoreProvider.PINECONE,
        VectorStoreProvider.MILVUS,
        VectorStoreProvider.QDRANT,
        VectorStoreProvider.CHROMA,
        VectorStoreProvider.WEAVIATE,
    ],
)
@pytest.mark.parametrize(
    "async_call",
    [False, True],
)
async def test_query(provider_type: str, async_call: bool):
    client = VectorStoreSyncAndAsyncClient(
        provider_type=provider_type, async_call=async_call
    )
    await create_collection_if_needed(provider_type, client)
    for item in vectors:
        await cleanup_vector(provider_type, get_key(item), client=client)

    batch = VectorBatch()
    for item in vectors:
        batch.put(
            key=get_key(item),
            value=item["value"],
            metadata=item["metadata"],
        )
    response = await client.batch(batch=batch)
    result = response.result
    await wait_for_indexing(provider_type)

    response = await client.query()
    result = response.result
    assert_select_result(provider_type, result.items, vectors, False)

    if provider_type not in except_count_providers:
        response = await client.count()
        result = response.result
        assert_count_result(result, len(vectors))

    for query in queries:
        if "except_providers" in query:
            if provider_type in query["except_providers"]:
                continue
        args = query["args"]
        filtered_items = filter_items(vectors, query["result_index"])

        projected = None
        if "select" in query["args"]:
            projected = query["args"]["select"]

        response = await client.query(**args)
        result = response.result
        ordered = True if "ordered" not in query else query["ordered"]
        assert_select_result(
            provider_type, result.items, filtered_items, ordered, projected
        )

        # count = query["count"]
        # result = await client.count(**args)
        # assert_count_result(result, count)

    batch = VectorBatch()
    for item in vectors:
        batch.delete(key=get_key(item))
    response = await client.batch(batch=batch)
    result = response.result

    await client.close()


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "provider_type",
    [
        VectorStoreProvider.PINECONE,
        VectorStoreProvider.MILVUS,
        VectorStoreProvider.QDRANT,
        VectorStoreProvider.CHROMA,
        VectorStoreProvider.WEAVIATE,
    ],
)
@pytest.mark.parametrize(
    "async_call",
    [False, True],
)
async def test_put_get_update_delete(provider_type: str, async_call: bool):
    vector = vectors[0]
    old_vector = vectors[0]
    new_vector = vectors[1]
    key = get_key(vector)
    client = VectorStoreSyncAndAsyncClient(
        provider_type=provider_type, async_call=async_call
    )
    await create_collection_if_needed(provider_type, client)
    await cleanup_vector(provider_type, key, client=client)

    with pytest.raises(NotFoundError):
        await client.get(key=key)

    response = await client.put(
        key=key, value=vector["value"], metadata=vector["metadata"]
    )
    result = response.result
    assert_put_result(result, vector)
    await wait_for_put(provider_type)

    response = await client.get(key=key)
    result = response.result
    assert_vector_item(provider_type, result, vector)

    vector["value"] = new_vector["value"]
    vector["metadata"] = new_vector["metadata"]
    response = await client.put(
        key=key, value=vector["value"], metadata=vector["metadata"]
    )
    result = response.result
    assert_put_result(result, vector)
    await wait_for_put(provider_type)

    response = await client.get(key=key)
    result = response.result
    assert_vector_item(provider_type, result, vector)

    if provider_type not in except_update_providers:
        vector["value"] = old_vector["value"]
        response = await client.update(key=key, value=vector["value"])
        result = response.result
        assert_put_result(result, vector)

        response = await client.get(key=key)
        result = response.result
        assert_vector_item(provider_type, result, vector)

        vector["metadata"] = old_vector["metadata"]
        response = await client.update_metadata(
            key=key, metadata=vector["metadata"]
        )
        result = response.result
        assert_put_result(result, vector)

        response = await client.get(key=key)
        result = response.result
        assert_vector_item(provider_type, result, vector)

    response = await client.delete(key=key)
    result = response.result
    assert result is None
    await wait_for_delete(provider_type)

    with pytest.raises(NotFoundError):
        await client.get(key=key)

    # delete is best effort
    await client.delete(key=key)

    await client.close()


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "provider_type",
    [
        VectorStoreProvider.PINECONE,
        VectorStoreProvider.MILVUS,
        VectorStoreProvider.QDRANT,
        VectorStoreProvider.CHROMA,
        VectorStoreProvider.WEAVIATE,
    ],
)
@pytest.mark.parametrize(
    "async_call",
    [False, True],
)
async def test_batch(provider_type: str, async_call: bool):
    client = VectorStoreSyncAndAsyncClient(
        provider_type=provider_type, async_call=async_call
    )
    await create_collection_if_needed(provider_type, client)
    for item in vectors:
        await cleanup_vector(provider_type, get_key(item), client=client)

    batch = VectorBatch()
    for item in vectors:
        batch.put(
            key=get_key(item),
            value=item["value"],
            metadata=item["metadata"],
        )
    response = await client.batch(batch=batch)
    result = response.result
    for i in range(0, len(vectors)):
        assert_put_result(result[i], vectors[i])

    await wait_for_put(provider_type)

    batch = VectorBatch()
    for item in vectors:
        batch.get(key=get_key(item))
    response = await client.batch(batch=batch)
    result = response.result
    result = sorted(result, key=lambda x: x.key.id)
    sorted_data = sorted(vectors, key=lambda x: x["id"])
    for i in range(0, len(vectors)):
        assert_vector_item(provider_type, result[i], sorted_data[i])

    batch = VectorBatch()
    for item in vectors:
        batch.delete(key=get_key(item))
    response = await client.batch(batch=batch)
    result = response.result
    for i in range(0, len(vectors)):
        assert result[i] is None

    await wait_for_delete(provider_type)
    for item in vectors:
        with pytest.raises(NotFoundError):
            await client.get(key=get_key(item))

    await client.close()


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "provider_type",
    [
        VectorStoreProvider.PINECONE,
        VectorStoreProvider.MILVUS,
        VectorStoreProvider.QDRANT,
        VectorStoreProvider.CHROMA,
        VectorStoreProvider.WEAVIATE,
    ],
)
@pytest.mark.parametrize(
    "async_call",
    [False, True],
)
async def test_delete(provider_type: str, async_call: bool):
    client = VectorStoreSyncAndAsyncClient(
        provider_type=provider_type, async_call=async_call
    )
    await create_collection_if_needed(provider_type, client)
    for item in vectors:
        await cleanup_vector(provider_type, get_key(item), client=client)

    batch = VectorBatch()
    for item in vectors:
        batch.put(
            key=get_key(item),
            value=item["value"],
            metadata=item["metadata"],
        )
    await client.batch(batch=batch)
    await wait_for_put(provider_type)

    for i in range(0, len(vectors)):
        await client.get(get_key(item))

    if provider_type not in except_delete_filter_providers:
        await client.delete(where="$metadata.bool = true")
        await wait_for_delete_bulk(provider_type)
        for i in range(0, 5):
            with pytest.raises(NotFoundError):
                await client.get(get_key(vectors[i]))

    for i in range(5, 10):
        await client.get(get_key(item))

    if provider_type not in except_delete_all_providers:
        await client.delete()
        await wait_for_delete_bulk(provider_type)
        for i in range(0, len(vectors)):
            with pytest.raises(NotFoundError):
                await client.get(get_key(vectors[i]))

    await client.close()


def get_key(vector):
    return vector["id"]


def assert_put_result(item: VectorItem, vector: dict):
    assert item.key.id == vector["id"]


def assert_vector_list(
    provider_type, result: list[VectorItem], sorted_index: list[int]
):
    assert len(result) == len(sorted_index)
    for i in range(len(sorted_index)):
        assert_vector_item(provider_type, result[i], vectors[sorted_index[i]])


def assert_vector_item(provider_type, item: VectorItem, vector: dict):
    assert item.key.id == vector["id"]
    if "value" in vector:
        for idx, val in enumerate(item.value.vector):
            assert round(val, 1) == vector["value"]["vector"][idx]
        if provider_type not in except_sparse_vector_providers:
            assert item.value.sparse_vector == vector["value"]["sparse_vector"]
        assert item.value.content == vector["value"]["content"]
    if "metadata" in vector:
        citem = None
        if vector["metadata"] is not None:
            citem = {}
            for k, v in vector["metadata"].items():
                if isinstance(v, list):
                    if provider_type not in except_metadata_list_providers:
                        citem[k] = v
                else:
                    citem[k] = v
        assert item.metadata == citem


def assert_select_result(
    provider_type: str,
    result: list,
    items: list,
    ordered: bool = True,
    projected: str | None = None,
):
    if projected is not None and projected != "*":
        fields = [field.strip() for field in projected.split(",")]
        projected_items = []
        for item in items:
            pitem = {}
            pitem["id"] = item["id"]
            for field in fields:
                pitem[field.replace("$", "")] = item[field.replace("$", "")]
            projected_items.append(pitem)
        items = projected_items
    assert len(result) == len(items)
    if not ordered:
        result = sorted(result, key=lambda x: x.key.id)
        items = sorted(items, key=lambda x: x["id"])
    for i in range(0, len(result)):
        assert_vector_item(provider_type, result[i], items[i])


def assert_count_result(result: int, count: int):
    assert result == count


def filter_items(items: list, index: list) -> list:
    result = []
    for i in range(0, len(index)):
        result.append(items[index[i]])
    return result


async def cleanup_vector(
    provider_type, key, client: VectorStoreSyncAndAsyncClient
):
    try:
        await client.delete(key=key)
    except NotFoundError:
        return

    # await wait_for_delete(provider_type)


async def wait_for_delete_bulk(provider_type):
    if provider_type == VectorStoreProvider.PINECONE:
        time.sleep(60)
    if provider_type == VectorStoreProvider.MILVUS:
        time.sleep(20)


async def wait_for_indexing(provider_type):
    if provider_type == VectorStoreProvider.PINECONE:
        time.sleep(20)
    if provider_type == VectorStoreProvider.MILVUS:
        time.sleep(20)


async def wait_for_delete(provider_type):
    if (
        provider_type == VectorStoreProvider.MILVUS
        or provider_type == VectorStoreProvider.PINECONE
    ):
        time.sleep(5)


async def wait_for_put(provider_type):
    if (
        provider_type == VectorStoreProvider.MILVUS
        or provider_type == VectorStoreProvider.PINECONE
    ):
        time.sleep(5)


async def create_collection_if_needed(
    provider_type: str, client: VectorStoreSyncAndAsyncClient
):
    response = await client.list_collections()
    result = response.result
    if "test" in result:
        return
    await client.create_collection(
        collection="test",
        config={"vector_index": {"field": "$value"}},
    )



from x8.core import Component, operation


class MainService(Component):
    @operation()
    def do(self, a: int) -> int:
        raise NotImplementedError

    @operation()
    def health(self) -> str:
        return "OK"



from x8.storage.document_store import DocumentStore


def run():
    c = DocumentStore()
    c.create_collection(collection="test")
    c.put(key="x", value={"a": "b"}, collection="test")


if __name__ == "__main__":
    run()



from __future__ import annotations

import json
import time
import uuid
from typing import Any

import redis
import redis.asyncio as aredis

from x8._common.redis_provider import RedisProvider
from x8.core import Context, DataModel, NCall, Operation, Response
from x8.core.exceptions import BadRequestError, ConflictError, NotFoundError

from ._models import (
    MessageBatch,
    MessageItem,
    MessageKey,
    MessageProperties,
    MessagePullConfig,
    MessagePutConfig,
    MessageValueType,
    MessagingMode,
    MessagingOperation,
    QueueConfig,
    QueueInfo,
    SubscriptionConfig,
    SubscriptionInfo,
    TopicConfig,
    TopicInfo,
)
from ._operation_parser import MessagingOperationParser


class RedisBase(RedisProvider):
    mode: MessagingMode

    queue: str | None
    topic: str | None
    subscription: str | None

    visibility_timeout: int
    check_pending_interval: float
    nparams: dict[str, Any]

    _client: redis.Redis | None
    _aclient: aredis.Redis | None
    _op_converter: OperationConverter
    _result_converter: ResultConverter
    _worker_id: str
    _last_pending_check: float

    def __init__(
        self,
        mode: MessagingMode = MessagingMode.QUEUE,
        queue: str | None = None,
        topic: str | None = None,
        subscription: str | None = None,
        url: str | None = None,
        host: str | None = None,
        port: int | None = None,
        db: int = 0,
        username: str | None = None,
        password: str | None = None,
        options: dict | None = None,
        visibility_timeout: int = 30,
        check_pending_interval: float = 30,
        nparams: dict[str, Any] = dict(),
        **kwargs: Any,
    ):
        """Initialize.

        Args:
            mode:
                Messaging mode. Defaults to QUEUE.
            queue:
                Queue name.
            topic:
                Topic name.
            subscription:
                Subscription name.
            url:
                Redis URL.
            host:
                Redis host.
            port:
                Redis port.
            db:
                Redis database.
            username:
                Redis username.
            password:
                Redis password.
            options:
                Redis options.
            visibility_timeout:
                Visibility timeout in seconds. Defaults to 30.
            check_pending_interval:
                Interval to check the pending queue.
                Defaults to 30.
            nparams:
                Native parameters to Redis client.
        """

        self.mode = mode
        self.queue = queue
        self.topic = topic
        self.subscription = subscription
        self.visibility_timeout = visibility_timeout
        self.check_pending_interval = check_pending_interval
        self.nparams = nparams

        self._client = None
        self._aclient = None
        self._op_converter = OperationConverter()
        self._result_converter = ResultConverter()
        self._worker_id = str(uuid.uuid4())
        self._last_pending_check = 0

        if self.mode == MessagingMode.QUEUE:
            self.topic = self.queue
            self.subscription = self.queue
        RedisProvider.__init__(
            self,
            url=url,
            host=host,
            port=port,
            db=db,
            username=username,
            password=password,
            options=options,
            **kwargs,
        )

    def __setup__(self, context: Context | None = None) -> None:
        if self._client is not None:
            return

        self._client, _ = self._get_client_and_lib(decode_responses=False)

    async def __asetup__(self, context: Context | None = None) -> None:
        if self._aclient is not None:
            return

        self._aclient, self._alib = self._aget_client_and_lib(
            decode_responses=False
        )

    def _get_topic_name(
        self, op_parser: MessagingOperationParser
    ) -> str | None:
        if self.mode == MessagingMode.PUBSUB:
            name = (
                op_parser.get_topic() or self.topic or self.__component__.topic
            )
        else:
            name = (
                op_parser.get_queue() or self.queue or self.__component__.queue
            )
        return name

    def _get_subscription_name(
        self, op_parser: MessagingOperationParser
    ) -> str | None:
        name = (
            op_parser.get_subscription()
            or self.subscription
            or self.__component__.subscription
        )
        if name is None and self.mode == MessagingMode.QUEUE:
            name = self._get_topic_name(op_parser)
        return name

    def __run__(
        self,
        operation: Operation | None = None,
        context: Context | None = None,
        **kwargs,
    ) -> Any:
        self.__setup__(context=context)
        op_parser = MessagingOperationParser(operation)
        ncall = self._get_ncall(
            op_parser,
            self._client,
            ClientHelper(
                self._client,
                self._op_converter,
            ),
        )
        if ncall is None:
            return super().__run__(
                operation=operation,
                context=context,
                **kwargs,
            )
        nresult = ncall.invoke()
        result = self._convert_nresult(nresult, op_parser)
        return Response(result=result, native=dict(result=nresult, call=ncall))

    async def __arun__(
        self,
        operation: Operation | None = None,
        context: Context | None = None,
        **kwargs,
    ) -> Any:
        await self.__asetup__(context=context)
        op_parser = MessagingOperationParser(operation)
        ncall = self._get_ncall(
            op_parser,
            self._aclient,
            AsyncClientHelper(
                self._aclient,
                self._op_converter,
            ),
        )
        if ncall is None:
            return await super().__arun__(
                operation=operation,
                context=context,
                **kwargs,
            )
        nresult = await ncall.ainvoke()
        result = self._convert_nresult(nresult, op_parser)
        return Response(result=result, native=dict(result=nresult, call=ncall))

    def _get_ncall(
        self,
        op_parser: MessagingOperationParser,
        client: Any,
        client_helper: ClientHelper | AsyncClientHelper,
    ) -> NCall | None:
        call = None
        op_converter = self._op_converter
        nargs = op_parser.get_nargs()

        # CREATE QUEUE
        if op_parser.op_equals(MessagingOperation.CREATE_QUEUE):
            args: dict[str, Any] | list = {
                "queue": self._get_topic_name(op_parser),
                "config": op_parser.get_queue_config(),
                "where_exists": op_parser.get_where_exists(),
                "nargs": nargs,
            }
            call = NCall(
                client_helper.create_queue,
                args,
                None,
            )
        # UPDATE QUEUE
        elif op_parser.op_equals(MessagingOperation.UPDATE_QUEUE):
            args = {
                "queue": self._get_topic_name(op_parser),
                "config": op_parser.get_queue_config(),
                "nargs": nargs,
            }
            call = NCall(
                client_helper.update_queue,
                args,
                None,
            )
        # DROP QUEUE
        elif op_parser.op_equals(MessagingOperation.DROP_QUEUE):
            args = {
                "topic": self._get_topic_name(op_parser),
                "where_exists": op_parser.get_where_exists(),
                "nargs": nargs,
            }
            call = NCall(
                client_helper.drop_topic,
                args,
                None,
            )
        # HAS QUEUE
        elif op_parser.op_equals(MessagingOperation.HAS_QUEUE):
            args = {
                "queue": self._get_topic_name(op_parser),
                "nargs": nargs,
            }
            call = NCall(
                client_helper.has_queue,
                args,
                None,
            )
        # GET QUEUE
        elif op_parser.op_equals(MessagingOperation.GET_QUEUE):
            args = {
                "queue": self._get_topic_name(op_parser),
                "nargs": nargs,
            }
            call = NCall(
                client_helper.get_queue,
                args,
                None,
            )
        # LIST QUEUES
        elif op_parser.op_equals(MessagingOperation.LIST_QUEUES):
            args = {
                "nargs": nargs,
            }
            call = NCall(
                client_helper.list_topics,
                args,
                None,
            )
        # CREATE TOPIC
        elif op_parser.op_equals(MessagingOperation.CREATE_TOPIC):
            args = {
                "topic": self._get_topic_name(op_parser),
                "config": op_parser.get_topic_config(),
                "where_exists": op_parser.get_where_exists(),
                "nargs": nargs,
            }
            call = NCall(
                client_helper.create_topic,
                args,
                None,
            )
        # UPDATE TOPIC
        elif op_parser.op_equals(MessagingOperation.UPDATE_TOPIC):
            args = {
                "topic": self._get_topic_name(op_parser),
                "config": op_parser.get_topic_config(),
                "nargs": nargs,
            }
            call = NCall(
                client_helper.update_topic,
                args,
                None,
            )
        # DROP TOPIC
        elif op_parser.op_equals(MessagingOperation.DROP_TOPIC):
            args = {
                "topic": self._get_topic_name(op_parser),
                "where_exists": op_parser.get_where_exists(),
                "nargs": nargs,
            }
            call = NCall(
                client_helper.drop_topic,
                args,
                None,
            )
        # HAS TOPIC
        elif op_parser.op_equals(MessagingOperation.HAS_TOPIC):
            args = {
                "topic": self._get_topic_name(op_parser),
                "nargs": nargs,
            }
            call = NCall(
                client_helper.has_topic,
                args,
                None,
            )
        # GET TOPIC
        elif op_parser.op_equals(MessagingOperation.GET_TOPIC):
            args = {
                "topic": self._get_topic_name(op_parser),
                "nargs": nargs,
            }
            call = NCall(
                client_helper.get_topic,
                args,
                None,
            )
        # LIST TOPICS
        elif op_parser.op_equals(MessagingOperation.LIST_TOPICS):
            args = {
                "nargs": nargs,
            }
            call = NCall(
                client_helper.list_topics,
                args,
                None,
            )
        # CREATE SUBSCRIPTION
        elif op_parser.op_equals(MessagingOperation.CREATE_SUBSCRIPTION):
            args = {
                "topic": self._get_topic_name(op_parser),
                "subscription": self._get_subscription_name(op_parser),
                "config": op_parser.get_subscription_config(),
                "where_exists": op_parser.get_where_exists(),
                "nargs": nargs,
            }
            call = NCall(
                client_helper.create_subscription,
                args,
                None,
            )
        # UPDATE SUBSCRIPTION
        elif op_parser.op_equals(MessagingOperation.UPDATE_SUBSCRIPTION):
            args = {
                "topic": self._get_topic_name(op_parser),
                "subscription": self._get_subscription_name(op_parser),
                "config": op_parser.get_subscription_config(),
                "nargs": nargs,
            }
            call = NCall(
                client_helper.update_subscription,
                args,
                None,
            )
        # DROP SUBSCRIPTION
        elif op_parser.op_equals(MessagingOperation.DROP_SUBSCRIPTION):
            args = {
                "topic": self._get_topic_name(op_parser),
                "subscription": self._get_subscription_name(op_parser),
                "where_exists": op_parser.get_where_exists(),
                "nargs": nargs,
            }
            call = NCall(
                client_helper.drop_subscription,
                args,
                None,
            )
        # HAS SUBSCRIPTION
        elif op_parser.op_equals(MessagingOperation.HAS_SUBSCRIPTION):
            args = {
                "topic": self._get_topic_name(op_parser),
                "subscription": self._get_subscription_name(op_parser),
                "nargs": nargs,
            }
            call = NCall(
                client_helper.has_subscription,
                args,
                None,
            )
        # GET SUBSCRIPTION
        elif op_parser.op_equals(MessagingOperation.GET_SUBSCRIPTION):
            args = {
                "topic": self._get_topic_name(op_parser),
                "subscription": self._get_subscription_name(op_parser),
                "nargs": nargs,
            }
            call = NCall(
                client_helper.get_subscription,
                args,
                None,
            )
        # LIST SUBSCRIPTIONS
        elif op_parser.op_equals(MessagingOperation.LIST_SUBSCRIPTIONS):
            args = {
                "topic": self._get_topic_name(op_parser),
                "nargs": nargs,
            }
            call = NCall(
                client_helper.list_subscriptions,
                args,
                None,
            )
        # PUT
        elif op_parser.op_equals(MessagingOperation.PUT):
            args = op_converter.convert_put(
                self._get_topic_name(op_parser),
                op_parser.get_value(),
                op_parser.get_metadata(),
                op_parser.get_properties(),
                op_parser.get_put_config(),
            )
            call = NCall(
                client.xadd,
                args,
                None,
            )
        # BATCH
        elif op_parser.op_equals(MessagingOperation.BATCH):
            args = {
                "topic": self._get_topic_name(op_parser),
                "batch": op_parser.get_batch(),
                "nargs": nargs,
            }
            call = NCall(
                client_helper.batch,
                args,
                None,
            )
        # PULL
        elif op_parser.op_equals(MessagingOperation.PULL):
            current_time = time.time()
            check_pending = (
                current_time - self._last_pending_check
                > self.check_pending_interval
            )
            if check_pending:
                self._last_pending_check = current_time
            args = {
                "topic": self._get_topic_name(op_parser),
                "subscription": self._get_subscription_name(op_parser),
                "config": op_parser.get_pull_config(),
                "worker_id": self._worker_id,
                "visible_timeout": self.visibility_timeout,
                "check_pending": check_pending,
                "nargs": nargs,
            }
            call = NCall(
                client_helper.pull,
                args,
                None,
            )
        # ACK
        elif op_parser.op_equals(MessagingOperation.ACK):
            args = op_converter.convert_ack(
                self._get_topic_name(op_parser),
                self._get_subscription_name(op_parser),
                op_parser.get_key(),
            )
            call = NCall(
                client.xack,
                args,
                None,
            )
        # NACK
        elif op_parser.op_equals(MessagingOperation.NACK):
            args = {
                "topic": self._get_topic_name(op_parser),
                "subscription": self._get_subscription_name(op_parser),
                "key": op_parser.get_key(),
                "nargs": nargs,
            }
            call = NCall(
                client_helper.nack,
                args,
                None,
            )
        # EXTEND
        elif op_parser.op_equals(MessagingOperation.EXTEND):
            args = {
                "topic": self._get_topic_name(op_parser),
                "subscription": self._get_subscription_name(op_parser),
                "key": op_parser.get_key(),
                "timeout": op_parser.get_timeout(),
                "worker_id": self._worker_id,
                "nargs": nargs,
            }
            call = NCall(
                client_helper.extend,
                args,
                None,
            )
        # PURGE
        elif op_parser.op_equals(MessagingOperation.PURGE):
            args = {
                "topic": self._get_topic_name(op_parser),
                "subscription": self._get_subscription_name(op_parser),
                "nargs": nargs,
            }
            call = NCall(
                client_helper.purge,
                args,
                None,
            )
        # CLOSE
        elif op_parser.op_equals(MessagingOperation.CLOSE):
            args = {
                "topic": op_parser.get_topic() or op_parser.get_queue(),
                "nargs": nargs,
            }
            call = NCall(
                client_helper.close,
                args,
                None,
            )
        return call

    def _convert_nresult(
        self,
        nresult: Any,
        op_parser: MessagingOperationParser,
    ) -> Any:
        result_converter = self._result_converter
        result: Any = None

        # CREATE QUEUE
        if op_parser.op_equals(MessagingOperation.CREATE_QUEUE):
            result = None
        # UPDATE QUEUE
        elif op_parser.op_equals(MessagingOperation.UPDATE_QUEUE):
            result = None
        # DROP QUEUE
        elif op_parser.op_equals(MessagingOperation.DROP_QUEUE):
            result = None
        # HAS QUEUE
        elif op_parser.op_equals(MessagingOperation.HAS_QUEUE):
            result = nresult
        # GET QUEUE
        elif op_parser.op_equals(MessagingOperation.GET_QUEUE):
            result = nresult
        # LIST QUEUES
        elif op_parser.op_equals(MessagingOperation.LIST_QUEUES):
            result = nresult
        # CREATE TOPIC
        elif op_parser.op_equals(MessagingOperation.CREATE_TOPIC):
            result = None
        # UPDATE TOPIC
        elif op_parser.op_equals(MessagingOperation.UPDATE_TOPIC):
            result = None
        # DROP TOPIC
        elif op_parser.op_equals(MessagingOperation.DROP_TOPIC):
            result = None
        # HAS TOPIC
        elif op_parser.op_equals(MessagingOperation.HAS_TOPIC):
            result = nresult
        # GET TOPIC
        elif op_parser.op_equals(MessagingOperation.GET_TOPIC):
            result = nresult
        # LIST TOPICS
        elif op_parser.op_equals(MessagingOperation.LIST_TOPICS):
            result = nresult
        # CREATE SUBSCRIPTION
        elif op_parser.op_equals(MessagingOperation.CREATE_SUBSCRIPTION):
            result = None
        # UPDATE SUBSCRIPTION
        elif op_parser.op_equals(MessagingOperation.UPDATE_SUBSCRIPTION):
            result = None
        # DROP SUBSCRIPTION
        elif op_parser.op_equals(MessagingOperation.DROP_SUBSCRIPTION):
            result = None
        # HAS SUBSCRIPTION
        elif op_parser.op_equals(MessagingOperation.HAS_SUBSCRIPTION):
            result = nresult
        # GET SUBSCRIPTION
        elif op_parser.op_equals(MessagingOperation.GET_SUBSCRIPTION):
            result = nresult
        # LIST SUBSCRIPTIONS
        elif op_parser.op_equals(MessagingOperation.LIST_SUBSCRIPTIONS):
            result = nresult
        # PUT
        elif op_parser.op_equals(MessagingOperation.PUT):
            result = None
        # BATCH
        elif op_parser.op_equals(MessagingOperation.BATCH):
            result = None
        # PULL
        elif op_parser.op_equals(MessagingOperation.PULL):
            result = result_converter.convert_pull(nresult)
        # ACK
        elif op_parser.op_equals(MessagingOperation.ACK):
            result = None
        # NACK
        elif op_parser.op_equals(MessagingOperation.NACK):
            result = None
        # EXTEND
        elif op_parser.op_equals(MessagingOperation.EXTEND):
            result = MessageItem(key=op_parser.get_key())
        # PURGE
        elif op_parser.op_equals(MessagingOperation.PURGE):
            result = None
        # CLOSE
        elif op_parser.op_equals(MessagingOperation.CLOSE):
            result = None
        return result


class ClientHelper:
    client: Any
    op_converter: OperationConverter

    def __init__(
        self,
        client: Any,
        op_converter: OperationConverter,
    ):
        self.client = client
        self.op_converter = op_converter

    def create_queue(
        self,
        queue: str,
        config: QueueConfig | None,
        where_exists: bool | None,
        nargs: Any,
    ):
        try:
            self.client.xgroup_create(
                name=queue, groupname=queue, id="$", mkstream=True
            )
        except redis.exceptions.ResponseError as e:
            if "BUSYGROUP" in str(e):
                if where_exists is False:
                    raise ConflictError
            else:
                raise

    def update_queue(
        self,
        queue: str,
        config: QueueConfig | None,
        nargs: Any,
    ):
        if not self.client.exists(queue):
            raise NotFoundError

    def has_queue(
        self,
        queue: str,
        nargs: Any,
    ) -> bool:
        return self.client.exists(queue) > 0

    def get_queue(
        self,
        queue: str,
        nargs: Any,
    ) -> QueueInfo:
        if not self.client.exists(queue):
            raise NotFoundError
        groups = self.client.xinfo_groups(name=queue)
        group = next(
            (
                group
                for group in groups
                if group["name"].decode("utf-8") == queue
            ),
            None,
        )
        if not group:
            raise NotFoundError
        active_count = self.count(
            queue, group["last-delivered-id"].decode("utf-8")
        )
        return QueueInfo(
            name=queue,
            active_message_count=active_count,
            inflight_message_count=group["pending"] if group else None,
            config=QueueConfig(),
            nref=queue,
        )

    def create_topic(
        self,
        topic: str,
        config: TopicConfig | None,
        where_exists: bool | None,
        nargs: Any,
    ):
        if self.client.exists(topic):
            if where_exists is False:
                raise ConflictError
            return
        try:
            self.client.xgroup_create(
                name=topic, groupname="dummy", id="$", mkstream=True
            )
            self.client.xgroup_destroy(
                name=topic,
                groupname="dummy",
            )
        except redis.exceptions.ResponseError as e:
            if "BUSYGROUP" in str(e):
                if where_exists is False:
                    raise ConflictError
            else:
                raise

    def update_topic(
        self,
        topic: str,
        config: TopicConfig | None,
        nargs: Any,
    ):
        if not self.client.exists(topic):
            raise NotFoundError

    def drop_topic(
        self,
        topic: str,
        where_exists: bool | None,
        nargs: Any,
    ):
        deleted = self.client.delete(topic)
        if not deleted and where_exists is True:
            raise NotFoundError

    def has_topic(
        self,
        topic: str,
        nargs: Any,
    ) -> bool:
        return self.client.exists(topic) > 0

    def get_topic(
        self,
        topic: str,
        nargs: Any,
    ) -> TopicInfo:
        if not self.client.exists(topic):
            raise NotFoundError
        groups = self.client.xinfo_groups(name=topic)
        return TopicInfo(
            name=topic,
            subscription_count=len(groups),
            config=TopicConfig(),
            nref=topic,
        )

    def list_topics(self, nargs: Any):
        result = []
        cursor = 0
        while True:
            cursor, keys = self.client.scan(cursor=cursor, _type="stream")
            for key in keys:
                if key.startswith(b"__"):
                    continue
                result.append(key.decode("utf-8"))
            if cursor == 0:
                break
        return result

    def create_subscription(
        self,
        topic: str,
        subscription: str,
        config: SubscriptionConfig | None,
        where_exists: bool | None,
        nargs: Any,
    ):
        try:
            self.client.xgroup_create(
                name=topic, groupname=subscription, id="$", mkstream=True
            )
        except redis.exceptions.ResponseError as e:
            if "BUSYGROUP" in str(e):
                if where_exists is False:
                    raise ConflictError
            else:
                raise

    def update_subscription(
        self,
        topic: str,
        subscription: str,
        config: SubscriptionConfig | None,
        nargs: Any,
    ):
        if not self.client.exists(topic):
            raise NotFoundError
        groups = self.client.xinfo_groups(name=topic)
        if not any(
            group["name"].decode("utf-8") == subscription for group in groups
        ):
            raise NotFoundError

    def drop_subscription(
        self,
        topic: str,
        subscription: str,
        where_exists: bool | None,
        nargs: Any,
    ):
        deleted = self.client.xgroup_destroy(
            name=topic,
            groupname=subscription,
        )
        if not deleted and where_exists is True:
            raise NotFoundError

    def has_subscription(
        self,
        topic: str,
        subscription: str,
        nargs: Any,
    ) -> bool:
        groups = self.client.xinfo_groups(name=topic)
        return any(
            group["name"].decode("utf-8") == subscription for group in groups
        )

    def get_subscription(
        self,
        topic: str,
        subscription: str,
        nargs: Any,
    ) -> SubscriptionInfo:
        if not self.client.exists(topic):
            raise NotFoundError
        groups = self.client.xinfo_groups(name=topic)
        group = next(
            (
                group
                for group in groups
                if group["name"].decode("utf-8") == subscription
            ),
            None,
        )
        if not group:
            raise NotFoundError
        active_count = self.count(
            topic, group["last-delivered-id"].decode("utf-8")
        )
        return SubscriptionInfo(
            name=subscription,
            topic=topic,
            active_message_count=active_count,
            inflight_message_count=group["pending"],
            config=SubscriptionConfig(),
            nref=subscription,
        )

    def list_subscriptions(self, topic: str, nargs: Any):
        result = []
        groups = self.client.xinfo_groups(name=topic)
        for group in groups:
            result.append(group["name"].decode("utf-8"))
        return result

    def batch(
        self,
        topic: str,
        batch: MessageBatch,
        nargs: Any,
    ):
        for operation in batch.operations:
            op_parser = MessagingOperationParser(operation)
            args = self.op_converter.convert_put(
                topic,
                op_parser.get_value(),
                op_parser.get_metadata(),
                op_parser.get_properties(),
                op_parser.get_put_config(),
            )
            self.client.xadd(**args)

    def check_pending_messages(
        self,
        topic: str,
        subscription: str,
        config: MessagePullConfig | None,
        worker_id: str,
        visibility_timeout: float,
        nargs: Any,
    ):
        max_count = config.max_count if config and config.max_count else 1
        pending = self.client.xpending_range(
            topic, subscription, "-", "+", max_count
        )
        stale_ids = [
            entry["message_id"]
            for entry in pending
            if entry["idle"] > visibility_timeout * 1000
        ]
        if stale_ids:
            nresult = self.client.xclaim(
                topic, subscription, worker_id, visibility_timeout, *stale_ids
            )
            stream_key, msgs = nresult[0] if nresult else (None, [])
            return msgs
        return []

    def pull(
        self,
        topic: str,
        subscription: str,
        config: MessagePullConfig | None,
        worker_id: str,
        visible_timeout: float,
        check_pending: bool,
        nargs: Any,
    ):
        msgs = []
        if check_pending:
            msgs = self.check_pending_messages(
                topic, subscription, config, worker_id, visible_timeout, nargs
            )
        if len(msgs) > 0 and config and config.max_count:
            new_config = config.copy()
            new_config.max_count = config.max_count - len(msgs)
        else:
            new_config = config
        args = self.op_converter.convert_pull(
            topic,
            subscription,
            new_config,
            worker_id,
        )
        nresult = self.client.xreadgroup(**args)
        stream_key, msgs = nresult[0] if nresult else (None, [])
        return msgs

    def nack(
        self,
        topic: str,
        subscription: str,
        key: MessageKey,
        nargs: Any,
    ):
        fields = key.nref["fields"].copy()
        fields[b"origin_id"] = key.nref["origin_id"] or key.nref["id"]
        self.client.xadd(
            name=topic,
            fields=fields,
            id="*",
        )
        self.client.xack(
            topic,
            subscription,
            key.nref["id"],
        )

    def purge(
        self,
        topic: str,
        subscription: str,
        nargs: Any,
    ):
        self.client.xgroup_destroy(topic, subscription)
        self.client.xgroup_create(topic, subscription, id="$", mkstream=True)

    def extend(
        self,
        topic: str,
        subscription: str,
        key: MessageKey,
        timeout: int | None,
        worker_id: str,
        nargs: Any,
    ):
        args = self.op_converter.convert_extend(
            topic,
            subscription,
            key,
            timeout,
            worker_id,
        )
        return self.client.xclaim(**args)

    def count(self, topic: str, last_id: str, batch_size: int = 1000):
        count = 0
        while True:
            entries = self.client.xrange(
                topic, min=f"({last_id}", max="+", count=batch_size
            )
            if not entries:
                break
            count += len(entries)
            last_id = list(entries)[-1][0].decode()
        return count

    def close(self, topic: str | None, nargs: Any):
        pass


class AsyncClientHelper:
    client: Any
    op_converter: OperationConverter

    def __init__(
        self,
        client: Any,
        op_converter: OperationConverter,
    ):
        self.client = client
        self.op_converter = op_converter

    async def create_queue(
        self,
        queue: str,
        config: QueueConfig | None,
        where_exists: bool | None,
        nargs: Any,
    ):
        try:
            await self.client.xgroup_create(
                name=queue, groupname=queue, id="$", mkstream=True
            )
        except redis.exceptions.ResponseError as e:
            if "BUSYGROUP" in str(e):
                if where_exists is False:
                    raise ConflictError
            else:
                raise

    async def update_queue(
        self,
        queue: str,
        config: QueueConfig | None,
        nargs: Any,
    ):
        if not await self.client.exists(queue):
            raise NotFoundError

    async def has_queue(
        self,
        queue: str,
        nargs: Any,
    ) -> bool:
        return await self.client.exists(queue) > 0

    async def get_queue(
        self,
        queue: str,
        nargs: Any,
    ) -> QueueInfo:
        if not await self.client.exists(queue):
            raise NotFoundError
        groups = await self.client.xinfo_groups(name=queue)
        group = next(
            (
                group
                for group in groups
                if group["name"].decode("utf-8") == queue
            ),
            None,
        )
        if not group:
            raise NotFoundError
        active_count = await self.count(
            queue, group["last-delivered-id"].decode("utf-8")
        )
        return QueueInfo(
            name=queue,
            active_message_count=active_count,
            inflight_message_count=group["pending"] if group else None,
            config=QueueConfig(),
            nref=queue,
        )

    async def create_topic(
        self,
        topic: str,
        config: TopicConfig | None,
        where_exists: bool | None,
        nargs: Any,
    ):
        if await self.client.exists(topic):
            if where_exists is False:
                raise ConflictError
            return
        try:
            await self.client.xgroup_create(
                name=topic, groupname="dummy", id="$", mkstream=True
            )
            await self.client.xgroup_destroy(
                name=topic,
                groupname="dummy",
            )
        except redis.exceptions.ResponseError as e:
            if "BUSYGROUP" in str(e):
                if where_exists is False:
                    raise ConflictError
            else:
                raise

    async def update_topic(
        self,
        topic: str,
        config: TopicConfig | None,
        nargs: Any,
    ):
        if not await self.client.exists(topic):
            raise NotFoundError

    async def drop_topic(
        self,
        topic: str,
        where_exists: bool | None,
        nargs: Any,
    ):
        deleted = await self.client.delete(topic)
        if not deleted and where_exists is True:
            raise NotFoundError

    async def has_topic(
        self,
        topic: str,
        nargs: Any,
    ) -> bool:
        return await self.client.exists(topic) > 0

    async def get_topic(
        self,
        topic: str,
        nargs: Any,
    ) -> TopicInfo:
        if not await self.client.exists(topic):
            raise NotFoundError
        groups = await self.client.xinfo_groups(name=topic)
        return TopicInfo(
            name=topic,
            subscription_count=len(groups),
            config=TopicConfig(),
            nref=topic,
        )

    async def list_topics(self, nargs: Any):
        result = []
        cursor = 0
        while True:
            cursor, keys = await self.client.scan(
                cursor=cursor, _type="stream"
            )
            for key in keys:
                if key.startswith(b"__"):
                    continue
                result.append(key.decode("utf-8"))
            if cursor == 0:
                break
        return result

    async def create_subscription(
        self,
        topic: str,
        subscription: str,
        config: SubscriptionConfig | None,
        where_exists: bool | None,
        nargs: Any,
    ):
        try:
            await self.client.xgroup_create(
                name=topic, groupname=subscription, id="$", mkstream=True
            )
        except redis.exceptions.ResponseError as e:
            if "BUSYGROUP" in str(e):
                if where_exists is False:
                    raise ConflictError
            else:
                raise

    async def update_subscription(
        self,
        topic: str,
        subscription: str,
        config: SubscriptionConfig | None,
        nargs: Any,
    ):
        if not await self.client.exists(topic):
            raise NotFoundError
        groups = await self.client.xinfo_groups(name=topic)
        if not any(
            group["name"].decode("utf-8") == subscription for group in groups
        ):
            raise NotFoundError

    async def drop_subscription(
        self,
        topic: str,
        subscription: str,
        where_exists: bool | None,
        nargs: Any,
    ):
        deleted = await self.client.xgroup_destroy(
            name=topic,
            groupname=subscription,
        )
        if not deleted and where_exists is True:
            raise NotFoundError

    async def has_subscription(
        self,
        topic: str,
        subscription: str,
        nargs: Any,
    ):
        groups = await self.client.xinfo_groups(name=topic)
        return any(
            group["name"].decode("utf-8") == subscription for group in groups
        )

    async def get_subscription(
        self,
        topic: str,
        subscription: str,
        nargs: Any,
    ) -> SubscriptionInfo:
        if not await self.client.exists(topic):
            raise NotFoundError
        groups = await self.client.xinfo_groups(name=topic)
        group = next(
            (
                group
                for group in groups
                if group["name"].decode("utf-8") == subscription
            ),
            None,
        )
        if not group:
            raise NotFoundError
        active_count = await self.count(
            topic, group["last-delivered-id"].decode("utf-8")
        )
        return SubscriptionInfo(
            name=subscription,
            topic=topic,
            active_message_count=active_count,
            inflight_message_count=group["pending"],
            config=SubscriptionConfig(),
            nref=subscription,
        )

    async def list_subscriptions(self, topic: str, nargs: Any):
        result = []
        groups = await self.client.xinfo_groups(name=topic)
        for group in groups:
            result.append(group["name"].decode("utf-8"))
        return result

    async def batch(
        self,
        topic: str,
        batch: MessageBatch,
        nargs: Any,
    ):
        for operation in batch.operations:
            op_parser = MessagingOperationParser(operation)
            args = self.op_converter.convert_put(
                topic,
                op_parser.get_value(),
                op_parser.get_metadata(),
                op_parser.get_properties(),
                op_parser.get_put_config(),
            )
            await self.client.xadd(**args)

    async def check_pending_messages(
        self,
        topic: str,
        subscription: str,
        config: MessagePullConfig | None,
        worker_id: str,
        visible_timeout: float,
        nargs: Any,
    ):
        max_count = config.max_count if config and config.max_count else 1
        pending = await self.client.xpending_range(
            topic, subscription, "-", "+", max_count
        )
        stale_ids = [
            entry["message_id"]
            for entry in pending
            if entry["idle"] > visible_timeout * 1000
        ]
        if stale_ids:
            nresult = await self.client.xclaim(
                topic, subscription, worker_id, visible_timeout, *stale_ids
            )
            stream_key, msgs = nresult[0] if nresult else (None, [])
            return msgs
        return []

    async def pull(
        self,
        topic: str,
        subscription: str,
        config: MessagePullConfig | None,
        worker_id: str,
        visible_timeout: float,
        check_pending: bool,
        nargs: Any,
    ):
        msgs = []
        if check_pending:
            msgs = await self.check_pending_messages(
                topic, subscription, config, worker_id, visible_timeout, nargs
            )
        if len(msgs) > 0 and config and config.max_count:
            new_config = config.copy()
            new_config.max_count = config.max_count - len(msgs)
        else:
            new_config = config
        args = self.op_converter.convert_pull(
            topic,
            subscription,
            new_config,
            worker_id,
        )
        nresult = await self.client.xreadgroup(**args)
        stream_key, msgs = nresult[0] if nresult else (None, [])
        return msgs

    async def nack(
        self,
        topic: str,
        subscription: str,
        key: MessageKey,
        nargs: Any,
    ):
        fields = key.nref["fields"].copy()
        fields[b"origin_id"] = key.nref["origin_id"] or key.nref["id"]
        await self.client.xadd(
            name=topic,
            fields=fields,
            id="*",
        )
        await self.client.xack(
            topic,
            subscription,
            key.nref["id"],
        )

    async def purge(
        self,
        topic: str,
        subscription: str,
        nargs: Any,
    ):
        await self.client.xgroup_destroy(topic, subscription)
        await self.client.xgroup_create(
            topic, subscription, id="$", mkstream=True
        )

    async def extend(
        self,
        topic: str,
        subscription: str,
        key: MessageKey,
        timeout: int | None,
        worker_id: str,
        nargs: Any,
    ):
        args = self.op_converter.convert_extend(
            topic,
            subscription,
            key,
            timeout,
            worker_id,
        )
        return await self.client.xclaim(**args)

    async def count(self, topic: str, last_id: str, batch_size: int = 1000):
        count = 0
        while True:
            entries = await self.client.xrange(
                topic, min=f"({last_id}", max="+", count=batch_size
            )
            if not entries:
                break
            count += len(entries)
            last_id = list(entries)[-1][0].decode()
        return count

    async def close(self, topic: str | None, nargs: Any):
        if not topic:
            await self.client.aclose()


class OperationConverter:
    def convert_put(
        self,
        topic: str | None,
        value: MessageValueType,
        metadata: dict | None,
        properties: MessageProperties | None,
        config: MessagePutConfig | None,
    ) -> dict[str, Any]:
        if config is not None and config.delay is not None:
            raise BadRequestError("Delay is not supported for Redis Streams")
        fields: dict[str, Any] = {}
        if isinstance(value, bytes):
            fields["body"] = value
            content_type = "application/octet-stream"
        elif isinstance(value, str):
            fields["body"] = value
            content_type = "text/plain"
        elif isinstance(value, dict):
            fields["body"] = json.dumps(value)
            content_type = "application/json"
        elif isinstance(value, DataModel):
            fields["body"] = value.to_json()
            content_type = "application/json"
        else:
            raise BadRequestError("Message type not supported")

        if properties:
            if properties.message_id:
                fields["message_id"] = properties.message_id
            if properties.group_id:
                fields["group_id"] = properties.group_id
        fields["content_type"] = content_type

        if metadata:
            fields["metadata"] = json.dumps(metadata)

        return {
            "name": topic,
            "fields": fields,
            "id": "*",
        }

    def convert_pull(
        self,
        topic: str,
        subscription: str,
        config: MessagePullConfig | None,
        worker_id: str,
    ) -> dict[str, Any]:
        count = config.max_count if config and config.max_count else 1
        block_ms = 0
        if config and config.max_wait_time:
            block_ms = int(config.max_wait_time * 1000)
        return {
            "groupname": subscription,
            "consumername": worker_id,
            "streams": {topic: ">"},
            "count": count,
            "block": block_ms if block_ms > 0 else None,
        }

    def convert_ack(
        self,
        topic: str | None,
        subscription: str | None,
        key: MessageKey,
    ) -> list:
        return [topic, subscription, key.nref["id"]]

    def convert_extend(
        self,
        topic: str,
        subscription: str,
        key: MessageKey,
        timeout: int | None,
        worker_id: str,
    ) -> dict[str, Any]:
        return {
            "name": topic,
            "groupname": subscription,
            "consumername": worker_id,
            "min_idle_time": 0,
            "message_ids": [key.nref["id"]],
        }


class ResultConverter:
    def _convert_key(
        self,
        id: bytes,
        fields: dict[bytes, bytes],
        origin_id: bytes | None,
    ) -> MessageKey:
        return MessageKey(
            nref={
                "id": id,
                "fields": fields,
                "origin_id": origin_id,
            }
        )

    def _convert_message(
        self, id: bytes, fields: dict[bytes, bytes]
    ) -> MessageItem:
        def _b2s(b: bytes) -> str:
            return b.decode("utf-8")

        body_raw: Any = fields.get(b"body")
        content_type = _b2s(fields.get(b"content_type", b"text/plain"))
        metadata_raw = fields.get(b"metadata")
        fields.get(b"body", None)

        if content_type == "application/octet-stream":
            value: MessageValueType = body_raw
        elif content_type == "application/json":
            value = json.loads(body_raw)
        else:
            value = _b2s(body_raw)

        origin_id = fields.get(b"origin_id", None)
        metadata = json.loads(metadata_raw) if metadata_raw else None
        props = MessageProperties(
            message_id=_b2s(fields.get(b"message_id", b"")) or None,
            content_type=content_type,
            group_id=_b2s(fields.get(b"group_id", b"")) or None,
            enqueued_time=float(_b2s(origin_id or id).split("-")[0]) / 1000.0,
            delivery_count=None,
        )
        return MessageItem(
            key=self._convert_key(id, fields, origin_id),
            value=value,
            metadata=metadata,
            properties=props,
        )

    def convert_pull(self, nresult: Any) -> list[MessageItem]:
        if not nresult:
            return []
        result: list[MessageItem] = []
        for message_id, fields in nresult:
            result.append(self._convert_message(message_id, fields))
        return result



from x8.core import Provider
from x8.core.exceptions import BadRequestError


class RedisProvider(Provider):
    url: str | None
    host: str | None
    port: int | None
    db: int
    username: str | None
    password: str | None
    options: dict | None

    def __init__(
        self,
        url: str | None = None,
        host: str | None = None,
        port: int | None = None,
        db: int = 0,
        username: str | None = None,
        password: str | None = None,
        options: dict | None = None,
        **kwargs
    ):
        self.url = url
        self.host = host
        self.port = port
        self.db = db
        self.username = username
        self.password = password
        self.options = options
        super().__init__(**kwargs)

    def _get_client_and_lib(self, decode_responses: bool = True):
        import redis

        lib = redis
        roptions = self.options if self.options else {}
        if self.url is not None:
            client = redis.from_url(
                self.url, decode_responses=decode_responses, **roptions
            )
        elif self.host is not None and self.port is not None:
            client = redis.Redis(
                host=self.host,
                port=self.port,
                db=self.db,
                username=self.username,
                password=self.password,
                decode_responses=decode_responses,
                **roptions,
            )
        else:
            raise BadRequestError(
                "Redis initialization needs url or host and port"
            )
        return client, lib

    def _aget_client_and_lib(self, decode_responses: bool = True):
        import redis.asyncio as redis

        lib = redis
        roptions = self.options if self.options else {}
        if self.url is not None:
            client = redis.from_url(
                self.url, decode_responses=decode_responses, **roptions
            )
        elif self.host is not None and self.port is not None:
            client = redis.Redis(
                host=self.host,
                port=self.port,
                db=self.db,
                username=self.username,
                password=self.password,
                decode_responses=decode_responses,
                **roptions,
            )
        else:
            raise BadRequestError(
                "Redis initialization needs url or host and port"
            )
        return client, lib



from __future__ import annotations

from typing import Any

from x8.core import Response, operation
from x8.ql import Expression, OrderBy, Select
from x8.storage._common import CollectionResult, StoreComponent

from ._models import (
    VectorBatch,
    VectorCollectionConfig,
    VectorItem,
    VectorKey,
    VectorList,
    VectorTransaction,
    VectorValue,
)


class VectorStore(StoreComponent):
    collection: str | None

    def __init__(self, collection: str | None = None, **kwargs):
        """_summary_

        Args:
            collection:
                Collection name.
        """
        self.collection = collection
        super().__init__(**kwargs)

    @operation()
    def create_collection(
        self,
        collection: str | None = None,
        config: dict | VectorCollectionConfig | None = None,
        where: str | Expression | None = None,
        **kwargs: Any,
    ) -> Response[CollectionResult]:
        """Create collection.

        Args:
            collection:
                Collection name.
            config:
                Collection config.
            where:
                Condition expression.

        Returns:
            Collection operation result.
        """
        raise NotImplementedError

    @operation()
    def drop_collection(
        self,
        collection: str | None = None,
        where: str | Expression | None = None,
        **kwargs: Any,
    ) -> Response[CollectionResult]:
        """Drop collection.

        Args:
            collection:
                Collection name.
            where:
                Condition expression.

        Returns:
            Collection operation result.
        """
        raise NotImplementedError

    @operation()
    def list_collections(
        self,
        **kwargs: Any,
    ) -> Response[list[str]]:
        """List collections.

        Returns:
            List of collection names.
        """
        raise NotImplementedError

    @operation()
    def has_collection(
        self,
        collection: str | None = None,
        **kwargs: Any,
    ) -> Response[bool]:
        """Check if the collection exists.

        Args:
            collection:
                Collection name.

        Returns:
            A value indicating whether the collection exists.
        """
        raise NotImplementedError

    @operation()
    def get(
        self,
        key: str | dict | VectorKey,
        collection: str | None = None,
        **kwargs: Any,
    ) -> Response[VectorItem]:
        """Get vector.

        Args:
            key:
                Vector key.
            collection:
                Collection name.

        Returns:
            Vector item.

        Raises:
            NotFoundError:
                Vector not found.
        """
        raise NotImplementedError

    @operation()
    def put(
        self,
        key: str | dict | VectorKey,
        value: list[float] | dict | VectorValue,
        metadata: dict | None = None,
        collection: str | None = None,
        **kwargs: Any,
    ) -> Response[VectorItem]:
        """Put vector.

        Args:
            key:
                Vector key.
            value:
                Vector value.
            metadata:
                Custom metadata.
            collection:
                Collection name.

        Returns:
            Vector item.
        """
        raise NotImplementedError

    @operation()
    def update(
        self,
        key: str | dict | VectorKey,
        value: list[float] | dict | VectorValue,
        collection: str | None = None,
        **kwargs: Any,
    ) -> Response[VectorItem]:
        """Update vector value.

        Args:
            key:
                Vector key.
            value:
                Vector value.
            collection:
                Collection name.

        Returns:
            Vector item.
        """
        raise NotImplementedError

    @operation()
    def update_metadata(
        self,
        key: str | dict | VectorKey,
        metadata: dict | None,
        collection: str | None = None,
        **kwargs: Any,
    ) -> Response[VectorItem]:
        """Update vector metadata.

        Args:
            key:
                Vector key.
            metadata:
                Vector metadata.
            collection:
                Collection name.

        Returns:
            Vector item.
        """
        raise NotImplementedError

    @operation()
    def delete(
        self,
        key: str | dict | VectorKey | None = None,
        where: str | Expression | None = None,
        collection: str | None = None,
        **kwargs: Any,
    ) -> Response[None]:
        """Delete vector.

        Args:
            key:
                Vector key.
            where:
                Condition expression.
            collection:
                Collection name.

        Returns:
            None.

        Raises:
            NotFoundError:
                Vector not found.
            PreconditionFailedError:
                Condition failed.
        """
        raise NotImplementedError

    @operation()
    def query(
        self,
        search: str | Expression | None = None,
        select: str | Select | None = None,
        where: str | Expression | None = None,
        order_by: str | OrderBy | None = None,
        limit: int | None = None,
        offset: int | None = None,
        collection: str | None = None,
        **kwargs,
    ) -> Response[VectorList]:
        """Query vectors.

        Args:
            search:
                Search expression.
            select:
                Select expression.
            where:
                Condition expression.
            order_by:
                Order by expression.
            limit:
                Query limit.
            offset:
                Query offset.
            collection:
                Collection name.

        Returns:
            List of vectors.
        """
        raise NotImplementedError

    @operation()
    def count(
        self,
        search: str | Expression | None = None,
        where: str | Expression | None = None,
        collection: str | None = None,
        **kwargs,
    ) -> Response[int]:
        """Count vectors.

        Args:
            search:
                Search expression.
            where:
                Condition expression.
            collection:
                Collection name.

        Returns:
            Vector count.
        """
        raise NotImplementedError

    @operation()
    def batch(
        self,
        batch: dict | VectorBatch,
        **kwargs,
    ) -> Response[list[Any]]:
        """Batch operation.

        Args:
            batch:
                Vector batch.

        Returns:
            Batch operation results.
        """
        raise NotImplementedError

    @operation()
    def transact(
        self,
        transaction: dict | VectorTransaction,
        **kwargs,
    ) -> Response[list[Any]]:
        """Execute transaction.

        Args:
            transaction:
                Vector transaction.

        Returns:
            Transaction results.

        Raises:
            ConflictError:
                Transaction failed.
        """
        raise NotImplementedError

    @operation()
    def close(
        self,
        **kwargs: Any,
    ) -> Response[None]:
        """Close the sync client.

        Returns:
            None.
        """
        raise NotImplementedError

    @operation()
    async def acreate_collection(
        self,
        collection: str | None = None,
        config: dict | VectorCollectionConfig | None = None,
        **kwargs: Any,
    ) -> Response[CollectionResult]:
        """Create collection.

        Args:
            collection:
                Collection name.
            config:
                Collection config.

        Returns:
            Collection operation result.
        """
        raise NotImplementedError

    @operation()
    async def adrop_collection(
        self,
        collection: str | None = None,
        **kwargs: Any,
    ) -> Response[CollectionResult]:
        """Drop collection.

        Args:
            collection:
                Collection name.

        Returns:
            Collection operation result.
        """
        raise NotImplementedError

    @operation()
    async def alist_collections(
        self,
        **kwargs: Any,
    ) -> Response[list[str]]:
        """List collections.

        Returns:
            List of collection names.
        """
        raise NotImplementedError

    @operation()
    async def ahas_collection(
        self,
        collection: str | None = None,
        **kwargs: Any,
    ) -> Response[bool]:
        """Check if the collection exists.

        Args:
            collection:
                Collection name.

        Returns:
            A value indicating whether the collection exists.
        """
        raise NotImplementedError

    @operation()
    async def aget(
        self,
        key: str | dict | VectorKey,
        collection: str | None = None,
        **kwargs: Any,
    ) -> Response[VectorItem]:
        """Get vector.

        Args:
            key:
                Vector key.
            collection:
                Collection name.

        Returns:
            Vector item.

        Raises:
            NotFoundError:
                Vector not found.
        """
        raise NotImplementedError

    @operation()
    async def aput(
        self,
        key: str | dict | VectorKey,
        value: list[float] | dict | VectorValue,
        metadata: dict | None = None,
        collection: str | None = None,
        **kwargs: Any,
    ) -> Response[VectorItem]:
        """Put vector.

        Args:
            key:
                Vector key.
            value:
                Vector value.
            metadata:
                Custom metadata.
            collection:
                Collection name.

        Returns:
            Vector item.
        """
        raise NotImplementedError

    @operation()
    async def aupdate(
        self,
        key: str | dict | VectorKey,
        value: list[float] | dict | VectorValue,
        collection: str | None = None,
        **kwargs: Any,
    ) -> Response[VectorItem]:
        """Update vector value.

        Args:
            key:
                Vector key.
            value:
                Vector value.
            collection:
                Collection name.

        Returns:
            Vector item.
        """
        raise NotImplementedError

    @operation()
    async def aupdate_metadata(
        self,
        key: str | dict | VectorKey,
        metadata: dict | None,
        collection: str | None = None,
        **kwargs: Any,
    ) -> Response[VectorItem]:
        """Update vector metadata.

        Args:
            key:
                Vector key.
            metadata:
                Vector metadata.
            collection:
                Collection name.

        Returns:
            Vector item.
        """
        raise NotImplementedError

    @operation()
    async def adelete(
        self,
        key: str | dict | VectorKey | None = None,
        where: str | Expression | None = None,
        collection: str | None = None,
        **kwargs: Any,
    ) -> Response[None]:
        """Delete vector.

        Args:
            key:
                Vector key.
            where:
                Condition expression.
            collection:
                Collection name.

        Returns:
            None.

        Raises:
            NotFoundError:
                Vector not found.
            PreconditionFailedError:
                Condition failed.
        """
        raise NotImplementedError

    @operation()
    async def aquery(
        self,
        search: str | Expression | None = None,
        select: str | Select | None = None,
        where: str | Expression | None = None,
        order_by: str | OrderBy | None = None,
        limit: int | None = None,
        offset: int | None = None,
        collection: str | None = None,
        **kwargs,
    ) -> Response[VectorList]:
        """Query vectors.

        Args:
            search:
                Search expression.
            select:
                Select expression.
            where:
                Condition expression.
            order_by:
                Order by expression.
            limit:
                Query limit.
            offset:
                Query offset.
            collection:
                Collection name.

        Returns:
            List of vectors.
        """
        raise NotImplementedError

    @operation()
    async def acount(
        self,
        search: str | Expression | None = None,
        where: str | Expression | None = None,
        collection: str | None = None,
        **kwargs,
    ) -> Response[int]:
        """Count vectors.

        Args:
            search:
                Search expression.
            where:
                Condition expression.
            collection:
                Collection name.

        Returns:
            Vector count.
        """
        raise NotImplementedError

    @operation()
    async def abatch(
        self,
        batch: dict | VectorBatch,
        **kwargs: Any,
    ) -> Response[list[Any]]:
        """Batch operation.

        Args:
            batch:
                Vector batch.

        Returns:
            Batch operation results.
        """
        raise NotImplementedError

    @operation()
    async def atransact(
        self,
        transaction: dict | VectorTransaction,
        **kwargs: Any,
    ) -> Response[list[Any]]:
        """Execute transaction.

        Args:
            transaction:
                Vector transaction.

        Returns:
            Transaction results.

        Raises:
            ConflictError:
                Transaction failed.
        """
        raise NotImplementedError

    @operation()
    async def aclose(
        self,
        **kwargs: Any,
    ) -> Response[None]:
        """Close the async client.

        Returns:
            None.
        """
        raise NotImplementedError
